'''
@author: Sascha Hornauer
This code is a heavily modified but based on Progressive Multi-Granularity Training of
Du, Ruoyi and Chang, Dongliang and Bhunia, Ayan Kumar and Xie, Jiyang and Song, Yi-Zhe and Ma, Zhanyu and Guo, Jun

See the LICENSE file for further citation information
'''

from __future__ import print_function
import os

import torch
import torch.optim as optim
import torch.nn as nn
import numpy as np
from lib.datasets import YearBuiltFolder
from torchvision.transforms import transforms
from lib.utils import load_model, cosine_anneal_schedule, jigsaw_generator, \
    test_softlabels, test
from torch.autograd.variable import Variable
import argparse

parser = argparse.ArgumentParser(description='Classify Year Built')

parser.add_argument('--image-path', help='Path to one image or a folder containing images.', required=True)
parser.add_argument('--epochs', help='Epochs to train', default=200)
parser.add_argument('--batch-size', help='Batch size to use for training. Recommended is 128 for 8 GPUs or significantly less when no GPU is available. Default is 8', default=8)
parser.add_argument('--soft-labels', help='Activate soft labels', action='store_true')
parser.add_argument('--gaussian-std', help='Standard deviation of the unimodal gaussian used to model class membership', default=1.5)
parser.add_argument('--checkpoint', default='', type=str,
                    help='Path to checkpoint. Defaults to best pretrained version.')
parser.add_argument('--exp-name', default='unnamed_experiment', help='Name of this experiment for saved files.')

args = parser.parse_args()

def train(nb_epoch, batch_size, store_name, image_path, soft_labels=False, gaussian_std=1.5, resume=False, start_epoch=0, model_path=None):
    batch_size = int(batch_size)
    nb_epoch = int(nb_epoch)
    start_epoch = int(start_epoch)
    
    exp_dir = store_name
    try:
        os.stat(exp_dir)
    except:
        os.makedirs(exp_dir)

    use_cuda = torch.cuda.is_available()
    # GPU
    # Data
    print('==> Preparing data..')
    transform_train = transforms.Compose([
        transforms.Scale((550, 550)),
        transforms.RandomCrop(448, padding=8),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor(),
        transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
    ])
    
    transform_test = transforms.Compose([
        transforms.Scale((550, 550)),
        transforms.CenterCrop(448),
        transforms.ToTensor(),
        transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
    ])
        
    trainset = YearBuiltFolder(image_path, calc_perf=True,soft_labels=soft_labels, gaussian_std=gaussian_std, transforms=transform_train)
    
    train_weights = np.array(trainset.train_weights)
    train_sampler = torch.utils.data.WeightedRandomSampler(train_weights, len(train_weights),
                                                                   replacement=True)
        
        
    if use_cuda:
        device = 'cuda'
    else:
        device = 'cpu'
            
    trainloader = torch.utils.data.DataLoader(trainset, batch_size=batch_size, shuffle=(train_sampler == None), num_workers=4, sampler=train_sampler)
    
    # Testing is on the same data but without the transforms. This whole class
    # only givey training performance and the user has to use the detect script
    # to check the performance.
    train_set_without_transforms = YearBuiltFolder(image_path,calc_perf=True, soft_labels=soft_labels, gaussian_std=gaussian_std, transforms=transform_test)
    
    train_test_loader = torch.utils.data.DataLoader(train_set_without_transforms, batch_size=3, shuffle=True, num_workers=4)
       
    # Model
    if resume:
        net = torch.load(model_path)
    else:
        net = load_model(model_name='resnet50_pmg', pretrain=True, require_grad=True, num_classes = len(trainset.classes))
    netp = torch.nn.DataParallel(net)

    net.to(device)
        
    if soft_labels:
        loss = nn.KLDivLoss(reduction='batchmean')
    else:
        loss = nn.CrossEntropyLoss()
    
    sm = nn.LogSoftmax(dim=1)
    
    optimizer = optim.SGD([
        {'params': net.classifier_concat.parameters(), 'lr': 0.002},
        {'params': net.conv_block1.parameters(), 'lr': 0.002},
        {'params': net.classifier1.parameters(), 'lr': 0.002},
        {'params': net.conv_block2.parameters(), 'lr': 0.002},
        {'params': net.classifier2.parameters(), 'lr': 0.002},
        {'params': net.conv_block3.parameters(), 'lr': 0.002},
        {'params': net.classifier3.parameters(), 'lr': 0.002},
        {'params': net.features.parameters(), 'lr': 0.0002}

    ],
        momentum=0.9, weight_decay=5e-4)

    max_train_acc = 0
    lr = [0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.002, 0.0002]
    for epoch in range(start_epoch, nb_epoch):
        print('\nEpoch: %d' % epoch)
        net.train()
        train_loss = 0
        train_loss1 = 0
        train_loss2 = 0
        train_loss3 = 0
        train_loss4 = 0
        correct = 0
        total = 0
        idx = 0
        for batch_idx, (inputs, targets, indexes) in enumerate(trainloader):
            idx = batch_idx
            if inputs.shape[0] < batch_size:
                continue
            if use_cuda:
                inputs, targets = inputs.to(device), targets.to(device)
            inputs, targets = Variable(inputs), Variable(targets)

            # update learning rate
            for nlr in range(len(optimizer.param_groups)):
                optimizer.param_groups[nlr]['lr'] = cosine_anneal_schedule(epoch, nb_epoch, lr[nlr])

            # Step 1
            optimizer.zero_grad()
            inputs1 = jigsaw_generator(inputs, 8)            
            output_1, _, _, _ = netp(inputs1)

            loss1 = loss(sm(output_1), targets) * 1
            loss1.backward()
            optimizer.step()

            # Step 2
            optimizer.zero_grad()
            inputs2 = jigsaw_generator(inputs, 4)
            _, output_2, _, _ = netp(inputs2)
            loss2 = loss(sm(output_2), targets) * 1
            loss2.backward()
            optimizer.step()

            # Step 3
            optimizer.zero_grad()
            inputs3 = jigsaw_generator(inputs, 2)
            _, _, output_3, _ = netp(inputs3)
            loss3 = loss(sm(output_3), targets) * 1
            loss3.backward()
            optimizer.step()

            # Step 4
            optimizer.zero_grad()
            _, _, _, output_concat = netp(inputs)
            concat_loss = loss(sm(output_concat), targets) * 2
            concat_loss.backward()
            optimizer.step()

            #  training log
            _, predicted = torch.max(output_concat.data, 1)
            if soft_labels:
                _, targets = torch.max(targets.data, 1)
            
            total += targets.size(0)
            correct += predicted.eq(targets.data).cpu().sum()

            train_loss += (loss1.item() + loss2.item() + loss3.item() + concat_loss.item())
            train_loss1 += loss1.item()
            train_loss2 += loss2.item()
            train_loss3 += loss3.item()
            train_loss4 += concat_loss.item()

            if batch_idx % 50 == 0:
                print(
                    'Step: %d | Loss1: %.3f | Loss2: %.5f | Loss3: %.5f | Loss_concat: %.5f | Loss: %.3f | Acc: %.3f%% (%d/%d)' % (
                    batch_idx, train_loss1 / (batch_idx + 1), train_loss2 / (batch_idx + 1),
                    train_loss3 / (batch_idx + 1), train_loss4 / (batch_idx + 1), train_loss / (batch_idx + 1),
                    100. * float(correct) / total, correct, total))

        train_acc = 100. * float(correct) / total
        train_loss = train_loss / (idx + 1)
        with open(os.path.join(exp_dir,'results_train.txt'), 'a') as file:
            file.write(
                'Iteration %d | train_acc = %.5f | train_loss = %.5f | Loss1: %.3f | Loss2: %.5f | Loss3: %.5f | Loss_concat: %.5f |\n' % (
                epoch, train_acc, train_loss, train_loss1 / (idx + 1), train_loss2 / (idx + 1), train_loss3 / (idx + 1),
                train_loss4 / (idx + 1)))

        # if epoch < 5 or epoch >= 80:
        print ('Start testing')
        if soft_labels:
            train_acc, train_acc_com, train_loss, y_gt, y_pred = test_softlabels(net, loss, train_test_loader)
        else:
            train_acc, train_acc_com, train_loss, y_gt, y_pred = test(net, loss, train_test_loader)
        if train_acc_com > max_train_acc:
            max_train_acc = train_acc_com
            net.cpu()
            torch.save(net, os.path.join(store_name,'model_best.pth'))
            net.to(device)
        with open(os.path.join(exp_dir,'results_test.txt'), 'a') as file:
            file.write('Iteration %d, test_acc = %.5f, test_acc_combined = %.5f, test_loss = %.6f\n' % (
            epoch, train_acc, train_acc_com, train_loss))
        
        net.cpu()
        torch.save(net, os.path.join(store_name,'model_epoch_{}.pth'.format(epoch)))
        net.to(device)


try:
    train(nb_epoch=args.epochs,  # number of epoch
             batch_size=args.batch_size,  # batch size
             image_path=args.image_path,
             soft_labels=args.soft_labels,
             gaussian_std=args.gaussian_std,
             store_name=args.exp_name,  # folder for output
             resume=not (args.checkpoint == ''),  # resume training from checkpoint
             start_epoch=0,  # the start epoch number when you resume the training
             model_path=args.checkpoint
             )  # the saved model where you want to resume the training
except ValueError as ex:
    if 'Expected more than 1 value per channel when training' in str(ex):
        print('Error: The chosen batch size is not compatible with the size of the dataset due to the batch norm layers in the model. Please choose a different batch size, preferably larger.')
        exit(1)



import os
import cv2
import numpy as np
from shapely.geometry import Polygon
from shapely.ops import unary_union
from lib.infer_detector import Infer
import torch
import time
from tqdm import tqdm
import warnings
import argparse

# Ignore Divide by Zero Warnings for dyOverdx
warnings.filterwarnings("ignore")

def get_args():
    parser = argparse.ArgumentParser('EfficientDet-based number of floor detection model')
    parser.add_argument('--im_path', type=str, default="datasets/test",
                        help='Path for the building images')
    parser.add_argument('--model_path', type=str, default="models/efficientdet-d4_trained.pth",
                        help='Path for the pretrained inference model.' 
                             'Do NOT define this argument if the pretrained model bundled with the module will be used')
    parser.add_argument('--gpu_enabled', type=boolean_string, default=True,
                        help='Enable GPU processing (Enter False for CPU-based inference)')    
    parser.add_argument('--csv_out', type=str, default="nFloorPredict.csv",
                        help='Name of the CSV output file where the inference results will be written')

    args = parser.parse_args()
    return args

def boolean_string(s):
    if s not in {'False', 'True'}:
        raise ValueError('Not a valid boolean entry')
    return s == 'True'

def install_default_model(model_path):
    if model_path == "models/efficientdet-d4_trained.pth":
        os.makedirs('models',exist_ok=True)
        model_path = os.path.join('models','efficientdet-d4_trained.pth')

        if not os.path.isfile(model_path):
            print('Loading default model file to the models folder...')
            torch.hub.download_url_to_file('https://zenodo.org/record/4421613/files/efficientdet-d4_trained.pth',
                                           model_path, progress=False)
            print('Default model loaded.')
    else:
        print(f'Inferences will be performed using the custom model at {model_path}.')

def create_polygon(bb):
    polygon = Polygon([(bb[0], bb[1]), (bb[0], bb[3]),
                       (bb[2], bb[3]), (bb[2], bb[1])])
    return polygon

def intersect_polygons(poly1, poly2):
    if poly1.intersects(poly2):
        polyArea = poly1.intersection(poly2).area
        if poly1.area!=0 and poly2.area!=0:
            overlapRatio = polyArea/poly1.area*100
        else: 
            overlapRatio = 0
    else:
        overlapRatio = 0
    return overlapRatio

def check_threshold_level(boxesPoly):
    thresholdChange = False
    thresholdIncrease = False
    if not boxesPoly:
        thresholdChange = True
        thresholdIncrease = False
    else:
        falseDetect = np.zeros(len(boxesPoly))
        for k in range(len(boxesPoly)):
            #for m in range(len(boxesPoly)):
            #    overlapRatio = np.array([intersect_polygons(boxesPoly[m], boxesPoly[k])],dtype=float)
            #    print(f"{m}, {k}")
            overlapRatio = np.array([intersect_polygons(p, boxesPoly[k]) for p in boxesPoly],dtype=float)
            falseDetect[k] = len([idx for idx, val in enumerate(overlapRatio) if val > 75][1:])
        thresholdChange = any(falseDetect>2)
        if thresholdChange: thresholdIncrease = True
    return thresholdChange, thresholdIncrease

def compute_derivative(centBoxes):
    nBoxes = centBoxes.shape[0] 
    dyOverdx = np.zeros((nBoxes,nBoxes)) + 10
    for k in range(nBoxes):
        for m in range(nBoxes):
            dx = abs(centBoxes[k,0]-centBoxes[m,0])
            dy = abs(centBoxes[k,1]-centBoxes[m,1])
            if k!=m:
                dyOverdx[k,m] = dy/dx
    return dyOverdx

def infer(opt):
    install_default_model(opt.model_path)
    
    # Start Program Timer
    startTime = time.time()
    
    # Get the Image List
    imgList = os.listdir(opt.im_path)
    nImages = len(imgList)
    
    # Initiate a CSV File to Write Program Output If More Than One Image in imageDir
    if nImages!=1:
        csvFile = open(opt.csv_out,'w+')
        csvFile.write("Image, nFloors\n")
        
    # Create and Define the Inference Model
    classes = ["floor"]
    
    print("Performing inferences on images...")
    gtfInfer = Infer()
    gtfInfer.load_model(opt.model_path, classes, use_gpu=opt.gpu_enabled)
    for imgno in tqdm(range(nImages)):
        # Perform Iterative Inference
        imgPath = os.path.join(opt.im_path, imgList[imgno])
        img = cv2.imread(imgPath)
        img = cv2.resize(img,(640,640))
        cv2.imwrite("input.jpg",img)
        _, _, boxes = gtfInfer.predict("input.jpg",threshold=0.2)
        boxesPoly = [create_polygon(bb) for bb in boxes]
        
        multiplier = 1
        while check_threshold_level(boxesPoly)[0]:
            if check_threshold_level(boxesPoly)[1]:
                confThreshold = 0.2 + multiplier*0.1
                if confThreshold>1: break
            else:
                confThreshold = 0.2 - multiplier*0.02
                if confThreshold==0: break
            _, _, boxes = gtfInfer.predict("input.jpg",threshold=confThreshold)                    
            multiplier +=1        
            boxesPoly = [create_polygon(bb) for bb in boxes]
        
        # Postprocessing    
        boxesPoly = [create_polygon(bb) for bb in boxes]
        
        nestedBoxes = np.zeros((10*len(boxes)),dtype=int)
        counter = 0
        for bbPoly in boxesPoly:    
            overlapRatio = np.array([intersect_polygons(p, bbPoly) for p in boxesPoly],dtype=float)
            ind = [idx for idx, val in enumerate(overlapRatio) if val > 75][1:]
            nestedBoxes[counter:counter+len(ind)] = ind
            counter += len(ind)
        nestedBoxes  = np.unique(nestedBoxes[:counter])
        
        counter = 0
        for x in nestedBoxes:
            del boxes[x-counter] 
            counter += 1
        
        nBoxes = len(boxes)
        
        boxesPoly = []
        boxesExtendedPoly = []
        centBoxes = np.zeros((nBoxes,2))
        for k in range(nBoxes):    
            bb = boxes[k]  
            tempPoly = create_polygon(bb)
            boxesPoly.append(tempPoly)
            x,y = tempPoly.centroid.xy
            centBoxes[k,:] = np.array([x[0],y[0]])
            boxesExtendedPoly.append(create_polygon([0.9*bb[0],0,1.1*bb[2],len(img)-1]))
        
        stackedInd = []
        for bb in boxesExtendedPoly:    
            overlapRatio = np.array([intersect_polygons(p, bb) for p in boxesExtendedPoly],dtype=float)
            stackedInd.append([idx for idx, val in enumerate(overlapRatio) if val > 10])
            
        uniqueStacks0 = [list(x) for x in set(tuple(x) for x in stackedInd)]
        
        dyOverdx = compute_derivative(centBoxes)
        stacks = np.where(dyOverdx>1.3)
        
        counter = 0
        uniqueStacks = [[] for i in range(nBoxes)]
        for k in range(nBoxes):
            while counter<len(stacks[0]) and k==stacks[0][counter]:
                uniqueStacks[k].append(stacks[1][counter])
                counter +=1
        
        uniqueStacks = [list(x) for x in set(tuple(x) for x in uniqueStacks)]
        
        if len(uniqueStacks0)==1 or len(uniqueStacks)==1:
            nFloors = len(uniqueStacks0[0])
        else:
            lBound = len(img)/5
            uBound = 4*len(img)/5
            middlePoly = Polygon([(lBound,0),(lBound,len(img)),(uBound,len(img)),(uBound,0)])
            overlapRatio = np.empty(len(uniqueStacks))
            for k in range(len(uniqueStacks)):
                poly = unary_union([boxesPoly[x] for x in uniqueStacks[k]])
                overlapRatio[k] = (intersect_polygons(poly, middlePoly))
        
            
            indKeep = np.argsort(-overlapRatio)[0:2]
            stack4Address = []
            for k in range(2):
                if overlapRatio[indKeep[k]]>10: stack4Address.append(indKeep[k])
            if len(stack4Address)!=0:
                nFloors = max([len(uniqueStacks[x]) for x in stack4Address])
            else:
                nFloors = len(uniqueStacks[0])
                
           
        if nImages!=1: 
                csvFile.write(os.path.splitext(imgList[imgno])[0] + f", {nFloors}\n")
        else:
            print(f"Number of floors:{nFloors}\n")        
    
    csvFile.close()
    # End Program Timer and Display Execution Time
    endTime = time.time()
    hours, rem = divmod(endTime-startTime, 3600)
    minutes, seconds = divmod(rem, 60)
    print("\nTotal execution time: {:0>2}:{:0>2}:{:05.2f}".format(int(hours),int(minutes),seconds))
    
    # Cleanup the Root Folder
    if os.path.isfile("input.jpg"):
        os.remove("input.jpg")
    if os.path.isfile("output.jpg"):
        os.remove("output.jpg")
        
if __name__ == '__main__':
    opt = get_args()
    infer(opt)


# -*- coding: utf-8 -*-
# File   : __init__.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 27/01/2018
# 
# This file is part of Synchronized-BatchNorm-PyTorch.
# https://github.com/vacancy/Synchronized-BatchNorm-PyTorch
# Distributed under MIT License.

from .batchnorm import SynchronizedBatchNorm1d, SynchronizedBatchNorm2d, SynchronizedBatchNorm3d
from .replicate import DataParallelWithCallback, patch_replication_callback



import torch.nn as nn
import math
from .utils import load_url
from csail_segmentation_tool.csail_seg.lib.nn import SynchronizedBatchNorm2d
BatchNorm2d = SynchronizedBatchNorm2d


__all__ = ['ResNeXt', 'resnext101'] # support resnext 101


model_urls = {
    #'resnext50': 'http://sceneparsing.csail.mit.edu/model/pretrained_resnet/resnext50-imagenet.pth',
    'resnext101': 'http://sceneparsing.csail.mit.edu/model/pretrained_resnet/resnext101-imagenet.pth'
}


def conv3x3(in_planes, out_planes, stride=1):
    "3x3 convolution with padding"
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)


class GroupBottleneck(nn.Module):
    expansion = 2

    def __init__(self, inplanes, planes, stride=1, groups=1, downsample=None):
        super(GroupBottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, groups=groups, bias=False)
        self.bn2 = BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 2, kernel_size=1, bias=False)
        self.bn3 = BatchNorm2d(planes * 2)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNeXt(nn.Module):

    def __init__(self, block, layers, groups=32, num_classes=1000):
        self.inplanes = 128
        super(ResNeXt, self).__init__()
        self.conv1 = conv3x3(3, 64, stride=2)
        self.bn1 = BatchNorm2d(64)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(64, 64)
        self.bn2 = BatchNorm2d(64)
        self.relu2 = nn.ReLU(inplace=True)
        self.conv3 = conv3x3(64, 128)
        self.bn3 = BatchNorm2d(128)
        self.relu3 = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.layer1 = self._make_layer(block, 128, layers[0], groups=groups)
        self.layer2 = self._make_layer(block, 256, layers[1], stride=2, groups=groups)
        self.layer3 = self._make_layer(block, 512, layers[2], stride=2, groups=groups)
        self.layer4 = self._make_layer(block, 1024, layers[3], stride=2, groups=groups)
        self.avgpool = nn.AvgPool2d(7, stride=1)
        self.fc = nn.Linear(1024 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels // m.groups
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1, groups=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, groups, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=groups))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.relu1(self.bn1(self.conv1(x)))
        x = self.relu2(self.bn2(self.conv2(x)))
        x = self.relu3(self.bn3(self.conv3(x)))
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return x


'''
def resnext50(pretrained=False, **kwargs):
    """Constructs a ResNet-50 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on Places
    """
    model = ResNeXt(GroupBottleneck, [3, 4, 6, 3], **kwargs)
    if pretrained:
        model.load_state_dict(load_url(model_urls['resnext50']), strict=False)
    return model
'''


def resnext101(pretrained=False, **kwargs):
    """Constructs a ResNet-101 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on Places
    """
    model = ResNeXt(GroupBottleneck, [3, 4, 23, 3], **kwargs)
    if pretrained:
        model.load_state_dict(load_url(model_urls['resnext101']), strict=False)
    return model


# def resnext152(pretrained=False, **kwargs):
#     """Constructs a ResNeXt-152 model.
#
#     Args:
#         pretrained (bool): If True, returns a model pre-trained on Places
#     """
#     model = ResNeXt(GroupBottleneck, [3, 8, 36, 3], **kwargs)
#     if pretrained:
#         model.load_state_dict(load_url(model_urls['resnext152']))
#     return model



import argparse

import torch
import torchvision.transforms as transforms
from torch.utils.data import DataLoader
import os
import csv

from PIL import Image
from torch.utils.data import Dataset
from torch.autograd.variable import Variable
import torch.nn as nn
import sklearn
#from sklearn.metrics.classification import precision_recall_fscore_support

import numpy as np
import pandas as pd

import matplotlib
matplotlib.use('agg')

import matplotlib.pyplot as plt
import seaborn as sns
from .lib.datasets import YearBuiltFolder


sm = nn.Softmax()

class YearBuiltClassifier():

    def __init__(self, checkpoint='', onlycpu=False, workDir='tmp', resultFile='YearBuilt.csv'):
        '''
        checkpoint (str): Path to checkpoint. Defaults to best pretrained version.
        onlycpu (bool): Use CPU only, use GPU by default.
        '''

        if onlycpu:
            self.device='cpu'
        else:
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
    
        self.checkpoint = checkpoint
        self.onlycpu = onlycpu
        self.workDir = workDir
        self.outFilePath = os.path.join(workDir, resultFile)

        self.checkpointsDir = os.path.join(workDir,'checkpoints')
        os.makedirs(self.checkpointsDir,exist_ok=True)
        weight_file_path = os.path.join(self.checkpointsDir,'yearBuiltv0.1.pth')

        if self.checkpoint != '':
            self.modelFile = self.checkpoint
            
        else:
            if not os.path.isfile(weight_file_path):
                print('Loading remote model file to the weights folder..')
                torch.hub.download_url_to_file('https://zenodo.org/record/4310463/files/model_best.pth', weight_file_path)
            self.modelFile = weight_file_path

        if not torch.cuda.is_available():
            self.model = torch.load(self.modelFile, map_location=torch.device('cpu'))
        else:
            self.model = torch.load(self.modelFile) 


        # test_transforms
        self.test_transforms = transforms.Compose([
            transforms.Scale((550, 550)),
            transforms.CenterCrop(448),
            transforms.ToTensor(),
            transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
        ])   

        try:
            self.num_classes = self.model.classifier1[4].out_features # Take amount of classes from model file
        except AttributeError:
            raise RuntimeError('The model, either provided or downloaded, does not fit the format in this implementation. Please check if the checkpoint is correct and fits to the model.')

        
        self.model.eval()
        self.model = self.model.to(self.device)

    def predict(self,image=''):
        '''
        image (str): Path to one image or a folder containing images.
        '''

        dataset = YearBuiltFolder(image,  transforms=self.test_transforms, classes=range(self.num_classes),calc_perf = False)
        # Do not change the batchsize.
        test_loader = DataLoader(dataset, batch_size=1, shuffle=False, num_workers=0) 

        predictions = []
        probs = []
        imagePathList = []

        predictions_data = self.evaluate_to_stats(test_loader)

        for prediction in predictions_data:
            imagePathList.append(str(prediction['filename']))
            predictions.append(prediction['prediction'][0])
            p = prediction['probability']
            probs.append(p)
            print(f"Image :  {str(prediction['filename'])}     Class : {prediction['prediction'][0]} ({str(round(p*100,2))}%)")

        df = pd.DataFrame(list(zip(imagePathList, predictions, probs)), columns =['image', 'prediction', 'probability']) 
        df.to_csv(self.outFilePath, index=False)
        print(f'Results written in file {self.outFilePath}')

        return df
        
    def evaluate_to_stats(self, testloader):
        self.model.eval()

        self.calc_perf = False

        predictions = []

        

        with torch.no_grad():
            for batch_idx, (inputs, label, filename) in enumerate(testloader):
            
                if inputs.shape[0] > 1:
                    raise NotImplementedError('Please choose a batch size of 1. Saving the results is not compatible with larger batch sizes in this version.')        
                inputs = inputs.to(self.device)
                inputs = Variable(inputs)
                output_1, output_2, output_3, output_concat= self.model(inputs)
                outputs_com = output_1 + output_2 + output_3 + output_concat

                
                output = sm(outputs_com.data)
                _, predicted_com = torch.max(output, 1)

                y_pred = predicted_com[0].flatten().cpu().numpy()

                p = output[0][y_pred][0].item()

                y_fn = filename[0]

                if self.calc_perf:
                    predictions.append({'filename':y_fn,'prediction':y_pred,'probability':p, 'ground truth':label.cpu().numpy()})
                else:
                    predictions.append({'filename':y_fn,'prediction':y_pred,'probability':p})


                if batch_idx % 50 == 0:
                    print('Testing image {} from {}'.format(batch_idx,len(testloader)))

            '''
            if self.calc_perf:
                y_gt = []
                y_pred = []
                for prediction in predictions:
                    y_gt.append(prediction['ground truth'])
                    y_pred.append(prediction['prediction'])

                y_gt = np.array(y_gt,dtype=np.uint8)
                y_pred = np.array(y_pred,dtype=np.uint8)

                precision, recall, f1, support = precision_recall_fscore_support(y_gt,
                                                                             y_pred, average='macro')

                confusion_matrix = sklearn.metrics.confusion_matrix(y_gt, y_pred, labels=range(len(testloader.dataset.classes)))

                cm_fig = construct_confusion_matrix_image(testloader.dataset.classes, confusion_matrix)

                cm_fig.savefig('result_confusion_matrix.png',dpi=300)

                print('F1 {}, precision, {}, recall, {}'.format(f1,precision,recall))   
            '''

        return predictions

    def construct_confusion_matrix_image(classes, con_mat):
        con_mat_norm = np.around(con_mat.astype('float') / con_mat.sum(axis=1)[:, np.newaxis], decimals=2)

        con_mat_df = pd.DataFrame(con_mat_norm,
                                  index=classes,
                                  columns=classes)

        figure = plt.figure(figsize=(8, 8))
        sns.heatmap(con_mat_df, annot=True, cmap=plt.cm.Blues)
        plt.tight_layout()
        plt.ylabel('True label')
        plt.xlabel('Predicted label')
        figure.canvas.draw()

        return figure



if __name__ == '__main__':
    main()



import argparse
import os

import shutil
import time

import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import torch.optim
import torch.utils.data
import torch.utils.data.distributed
import torchvision.transforms as transforms

from utils.Datasets import Foundation_Type_Binary
from lib.NCEAverage import NCEAverage
from lib.LinearAverage import LinearAverage
from lib.NCECriterion import NCECriterion
from lib.utils import AverageMeter
from npid.npid_models.resnet import resnet50
from npid.test import kNN, NN

parser = argparse.ArgumentParser(description='PyTorch ImageNet Training')
parser.add_argument('--train-data',type=str,
                    help='path to dataset')
parser.add_argument('--val-data', type=str,
                    help='path to dataset')
parser.add_argument('-j', '--workers', default=4, type=int, metavar='N',
                    help='number of data loading workers (default: 4)')
parser.add_argument('--epochs', default=200, type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('--start-epoch', default=0, type=int, metavar='N',
                    help='manual epoch number (useful on restarts)')
parser.add_argument('-b', '--batch-size', default=256, type=int,
                    metavar='N', help='mini-batch size (default: 256)')
parser.add_argument('--lr', '--learning-rate', default=0.03, type=float,
                    metavar='LR', help='initial learning rate')
parser.add_argument('--momentum', default=0.9, type=float, metavar='M',
                    help='momentum')
parser.add_argument('--weight-decay', '--wd', default=1e-4, type=float,
                    metavar='W', help='weight decay (default: 1e-4)')
parser.add_argument('--print-freq', '-p', default=10, type=int,
                    metavar='N', help='print frequency (default: 10)')
parser.add_argument('--resume', default='', type=str, metavar='PATH',
                    help='path to latest checkpoint (default: none)')
parser.add_argument('--test-only', action='store_true', help='test only')
parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true',
                    help='evaluate model on validation set')
parser.add_argument('--pretrained', dest='pretrained', action='store_true',
                    help='use pre-trained model')
parser.add_argument('--world-size', default=1, type=int,
                    help='number of distributed processes')
parser.add_argument('--dist-url', default='tcp://224.66.41.62:23456', type=str,
                    help='url used to set up distributed training')
parser.add_argument('--dist-backend', default='gloo', type=str,
                    help='distributed backend')
parser.add_argument('--low-dim', default=128, type=int,
                    metavar='D', help='feature dimension')
parser.add_argument('--nce-k', default=4096, type=int,
                    metavar='K', help='number of negative samples for NCE')
parser.add_argument('--nce-t', default=0.07, type=float, 
                    metavar='T', help='temperature parameter for softmax')
parser.add_argument('--nce-m', default=0.5, type=float,
                    help='momentum for non-parametric updates')
parser.add_argument('--iter_size', default=1, type=int,
                    help='caffe style iter size')
parser.add_argument('--finetune',action='store_true')
parser.add_argument('--name', type=str,default='unnamed')

parser.add_argument('--mask-buildings',action='store_true')

best_prec1 = 0

def main():
    global args, best_prec1
    args = parser.parse_args()

    args.distributed = args.world_size > 1

    if args.distributed:
        dist.init_process_group(backend=args.dist_backend, init_method=args.dist_url,
                                world_size=args.world_size)

    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])

    train_transform = transforms.Compose(
        [
            transforms.RandomResizedCrop(224, scale=(0.3, 1.)),
            transforms.RandomGrayscale(p=0.5),
            transforms.ColorJitter(0.5, 0.5, 0.5, 0.5),
            transforms.RandomHorizontalFlip(),
            transforms.RandomVerticalFlip(),
            transforms.ToTensor(),
            normalize])

    val_transform = transforms.Compose(
        [
            transforms.Resize((224, 224)),
            transforms.ToTensor(),
            normalize])

    train_dataset = Foundation_Type_Binary(args.train_data, transform=train_transform, mask_buildings=args.mask_buildings)
    val_dataset = Foundation_Type_Binary(args.val_data, transform=val_transform, mask_buildings=args.mask_buildings)

    train_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=args.batch_size, shuffle=True,
        num_workers=args.workers, pin_memory=True)

    val_loader = torch.utils.data.DataLoader(val_dataset,
        batch_size=args.batch_size, shuffle=False,
        num_workers=args.workers, pin_memory=True)

    model = resnet50(low_dim=args.low_dim)
    model = torch.nn.DataParallel(model).cuda()

    print ('Train dataset instances: {}'.format(len(train_loader.dataset)))
    print ('Val dataset instances: {}'.format(len(val_loader.dataset)))

    ndata = train_dataset.__len__()
    if args.nce_k > 0:
        lemniscate = NCEAverage(args.low_dim, ndata, args.nce_k, args.nce_t, args.nce_m).cuda()
        criterion = NCECriterion(ndata).cuda()
    else:
        lemniscate = LinearAverage(args.low_dim, ndata, args.nce_t, args.nce_m).cuda()
        criterion = nn.CrossEntropyLoss().cuda()

    optimizer = torch.optim.SGD(model.parameters(), args.lr,
                                 momentum=args.momentum,
                                 weight_decay=args.weight_decay)

    # optionally resume from a checkpoint
    if args.resume:
        if os.path.isfile(args.resume):
            print("=> loading checkpoint '{}'".format(args.resume))
            checkpoint = torch.load(args.resume)
            args.start_epoch = checkpoint['epoch']
            args.epochs = args.start_epoch + args.epochs
            best_prec1 = checkpoint['best_prec1']

            missing_keys, unexpected_keys = model.load_state_dict(checkpoint['state_dict'], strict=False)
            if len(missing_keys) or len(unexpected_keys):
                print('Warning: Missing or unexpected keys found.')
                print('Missing: {}'.format(missing_keys))
                print('Unexpected: {}'.format(unexpected_keys))

            low_dim_checkpoint = checkpoint['lemniscate'].memory.shape[1]
            if low_dim_checkpoint == args.low_dim:
                lemniscate = checkpoint['lemniscate']
            else:
                print('Chosen low dim does not fit checkpoint. Assuming fine-tuning and not loading memory bank.')
            try:
                optimizer.load_state_dict(checkpoint['optimizer'])
            except ValueError:
                print('Training optimizer does not fit to checkpoint optimizer. Assuming fine-tuning and load optimizer from scratch. ')

            print("=> loaded checkpoint '{}' (epoch {})"
                  .format(args.resume, checkpoint['epoch']))
        else:
            print("=> no checkpoint found at '{}'".format(args.resume))

    cudnn.benchmark = True

    if args.evaluate:
        kNN(0, model, lemniscate, train_loader, val_loader, 200, args.nce_t)
        return

    prec1 = NN(0, model, lemniscate, train_loader, val_loader)
    print('Start out precision: {}'.format(prec1))
    for epoch in range(args.start_epoch, args.epochs):

        adjust_learning_rate(optimizer, epoch)

        # train for one epoch
        train(train_loader, model, lemniscate, criterion, optimizer, epoch)

        # evaluate on validation set
        prec1 = NN(epoch, model, lemniscate, train_loader, val_loader)

        # remember best prec@1 and save checkpoint
        is_best = prec1 > best_prec1
        best_prec1 = max(prec1, best_prec1)
        save_checkpoint({
            'epoch': epoch + 1,
            'state_dict': model.state_dict(),
            'lemniscate': lemniscate,
            'best_prec1': best_prec1,
            'optimizer' : optimizer.state_dict(),
        }, is_best, args.name)

def train(train_loader, model, lemniscate, criterion, optimizer, epoch):
    batch_time = AverageMeter()
    data_time = AverageMeter()
    losses = AverageMeter()

    # switch to train mode
    model.train()

    end = time.time()
    optimizer.zero_grad()

    for i, (input, _, index) in enumerate(train_loader):
        # measure data loading time
        data_time.update(time.time() - end)

        index = index.cuda(non_blocking=True)

        # compute output
        feature = model(input)
        output = lemniscate(feature, index)
        loss = criterion(output, index) / args.iter_size

        loss.backward()

        # measure accuracy and record loss
        losses.update(loss.item() * args.iter_size, input.size(0))

        if (i+1) % args.iter_size == 0:
            # compute gradient and do SGD step
            optimizer.step()
            optimizer.zero_grad()

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % args.print_freq == 0:
            print('Epoch: [{0}][{1}/{2}]\t'
                  'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'
                  'Data {data_time.val:.3f} ({data_time.avg:.3f})\t'
                  'Loss {loss.val:.4f} ({loss.avg:.4f})\t'.format(
                   epoch, i, len(train_loader), batch_time=batch_time,
                   data_time=data_time, loss=losses))


def save_checkpoint(state, is_best, prefix, filename='checkpoint.pth.tar'):
    filename = prefix + '-' + filename
    torch.save(state, filename)
    if is_best:
        torch.save(state, prefix + '-model_best.pth.tar')

def adjust_learning_rate(optimizer, epoch):
    """Sets the learning rate to the initial LR decayed by 10 every 100 epochs"""
    lr = args.lr
    if epoch < 120:
        lr = args.lr
    elif epoch >= 120 and epoch < 160:
        lr = args.lr * 0.1
    else:
        lr = args.lr * 0.01

    for param_group in optimizer.param_groups:
        param_group['lr'] = lr


def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


if __name__ == '__main__':
    main()



import pandas as pd
import os
def generate(root_dir, is_train=True):
    df = pd.DataFrame(columns=['dir', 'class'])
    if(is_train):
        for d in range(1,7):
            cur_cat = str(d+5000)
            cur_dir = os.path.join(root_dir, cur_cat)
            for l in os.listdir(cur_dir):
                df.loc[len(df)] = [os.path.join(cur_dir, l), cur_cat]
        df.to_csv(os.path.join(root_dir, 'description_train.csv'))
    else:
        for d in range(1,7):
            cur_cat = str(d+5000)
            cur_dir = os.path.join(root_dir, cur_cat)
            for l in os.listdir(cur_dir):
                df.loc[len(df)] = [os.path.join(cur_dir, l), cur_cat]
        df.to_csv(os.path.join(root_dir, 'description_test.csv'))


generate('/data/sascha/Simcenter/cleaned_images/images/train')
generate('/data/sascha/Simcenter/cleaned_images/images/validate', is_train=False)
generate('/data/sascha/Simcenter/cleaned_images/test', is_train=False)


import argparse
import datetime
import os
import random
import shutil
import time
import warnings

import matplotlib
matplotlib.use('agg')

import numpy as np
import sklearn

import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.distributed as dist

import torch.multiprocessing as mp

import torch.utils.data.distributed
import torchvision.transforms as transforms
from sklearn.metrics import precision_recall_fscore_support
from tensorboardX import SummaryWriter

from .utils.datasets import RoofImages
import torchvision.models as models

from .utils.eval_tools import construct_confusion_matrix_image
from .utils.radam import RAdam
import pandas as pd

model_names = sorted(name for name in models.__dict__
    if name.islower() and not name.startswith("__")
    and callable(models.__dict__[name]))


class RoofMaterialClassifier():

    def __init__(self, checkpoint='', onlycpu=False, workDir='tmp', resultFile='YearBuilt.csv'):
        '''
        checkpoint (str): Path to checkpoint. Defaults to best pretrained version.
        onlycpu (bool): Use CPU only, use GPU by default.
        '''

        if onlycpu:
            self.device='cpu'
        else:
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
    
        self.checkpoint = checkpoint
        self.onlycpu = onlycpu
        self.workDir = workDir
        self.outFilePath = os.path.join(workDir, resultFile)

        self.checkpointsDir = os.path.join(workDir,'checkpoints')
        os.makedirs(self.checkpointsDir,exist_ok=True)
        weight_file_path = os.path.join(self.checkpointsDir,'weights_resnet_34.ckp')

        if self.checkpoint != '':
            self.modelFile = self.checkpoint
            
        else:
            if not os.path.isfile(weight_file_path):
                print('Loading remote model file to the weights folder..')
                torch.hub.download_url_to_file('https://zenodo.org/record/4394542/files/weights_resnet_34.ckp', weight_file_path)
            self.modelFile = weight_file_path
        
        self.train-data = ''
        self.val-data = imgDir

parser.add_argument('--train-data',
                    help='path to training data, either a csv file ending in .csv, a folder or a single image')
parser.add_argument('--val-data',
                    help='path to val csv, either a csv file ending in .csv, a folder or a single image. For testing, provide the test data with this flag and  choose the -e eval command line parameter')
parser.add_argument('-a', '--arch', metavar='ARCH', default='resnet34',
                    choices=model_names,
                    help='model architecture: ' +
                        ' | '.join(model_names) +
                        ' (default: resnet18)')
parser.add_argument('-j', '--workers', default=4, type=int, metavar='N',
                    help='number of data loading workers (default: 4)')
parser.add_argument('--epochs', default=90, type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('--start-epoch', default=0, type=int, metavar='N',
                    help='manual epoch number (useful on restarts)')
parser.add_argument('-b', '--batch-size', default=64, type=int,
                    metavar='N',
                    help='mini-batch size (default: 256), this is the total '
                         'batch size of all GPUs on the current node when '
                         'using Data Parallel or Distributed Data Parallel')
parser.add_argument('--lr', '--learning-rate', default=1e-3, type=float,
                    metavar='LR', help='initial learning rate', dest='lr')
parser.add_argument('--momentum', default=0.9, type=float, metavar='M',
                    help='momentum')
parser.add_argument('--wd', '--weight-decay', default=1e-4, type=float,
                    metavar='W', help='weight decay (default: 1e-4)',
                    dest='weight_decay')
parser.add_argument('-p', '--print-freq', default=10, type=int,
                    metavar='N', help='print frequency (default: 10)')
parser.add_argument('--resume', type=str, metavar='PATH',
                    help='path to latest checkpoint (default: none)')
parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true',
                    help='evaluate model on validation set')
parser.add_argument('--pretrained', dest='pretrained', action='store_true',
                    help='use pre-trained model')
parser.add_argument('--world-size', default=-1, type=int,
                    help='number of nodes for distributed training')
parser.add_argument('--rank', default=-1, type=int,
                    help='node rank for distributed training')
parser.add_argument('--dist-url', default='tcp://224.66.41.62:23456', type=str,
                    help='url used to set up distributed training')
parser.add_argument('--dist-backend', default='nccl', type=str,
                    help='distributed backend')
parser.add_argument('--seed', default=None, type=int,
                    help='seed for initializing training. ')
parser.add_argument('--gpu', default=None, type=int,
                    help='GPU id to use.')
parser.add_argument('--loss-weighting', action='store_true',
                    help='loss weighting.')
parser.add_argument('--val-resize', type = int, default=None,
                    help='validation image resizing')
parser.add_argument('--weighted-sampling', action='store_true')
parser.add_argument('--multiprocessing-distributed', action='store_true',
                    help='Use multi-processing distributed training to launch '
                         'N processes per node, which has N GPUs. This is the '
                         'fastest way to use PyTorch for either single node or '
                         'multi node data parallel training')
parser.add_argument('--checkpoint-dir', default='.')
parser.add_argument('--result-file', default='results.csv', help='Name of the csv file to save the predictions of all files.')

best_f1 = 0

os.makedirs('weights',exist_ok=True)
weight_file_path = os.path.join('weights','weights_resnet_34.ckp')

    def predict(self,image=''):
        '''
        image (str): Path to one image or a folder containing images.
        '''

        dataset = YearBuiltFolder(image,  transforms=self.test_transforms, classes=range(self.num_classes),calc_perf = False)
        # Do not change the batchsize.
        test_loader = DataLoader(dataset, batch_size=1, shuffle=False, num_workers=0) 

        predictions = []
        probs = []
        imagePathList = []

        predictions_data = self.evaluate_to_stats(test_loader)

        for prediction in predictions_data:
            imagePathList.append(str(prediction['filename']))
            predictions.append(prediction['prediction'][0])
            p = None
            probs.append(p)
            print(f"Image :  {str(prediction['filename'])}     Class : {prediction['prediction'][0]}")

        df = pd.DataFrame(list(zip(imagePathList, predictions, probs)), columns =['image', 'prediction', 'probability']) 
        df.to_csv(self.outFilePath, index=False)
        print(f'Results written in file {self.outFilePath}')

        return df
       
    def main():
        args = parser.parse_args()

        if args.evaluate and args.resume == None:
            args.resume = weight_file_path

        currentTime = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")

        args.log_dir = os.path.join('runs', 'run_{}'.format(args.checkpoint_dir) + currentTime)
        os.makedirs('runs', exist_ok=True)

        ## Tensorboard ##
        writer = SummaryWriter(args.log_dir)
        writer.add_text('args', str(args), 0)

        if args.seed is not None:
            random.seed(args.seed)
            torch.manual_seed(args.seed)
            cudnn.deterministic = True
            warnings.warn('You have chosen to seed training. '
                          'This will turn on the CUDNN deterministic setting, '
                          'which can slow down your training considerably! '
                          'You may see unexpected behavior when restarting '
                          'from checkpoints.')

        if args.gpu is not None:
            warnings.warn('You have chosen a specific GPU. This will completely '
                          'disable data parallelism.')

        if args.dist_url == "env://" and args.world_size == -1:
            args.world_size = int(os.environ["WORLD_SIZE"])

        args.distributed = args.world_size > 1 or args.multiprocessing_distributed

        ngpus_per_node = torch.cuda.device_count()
        if args.multiprocessing_distributed:
            # Since we have ngpus_per_node processes per node, the total world_size
            # needs to be adjusted accordingly
            args.world_size = ngpus_per_node * args.world_size
            # Use torch.multiprocessing.spawn to launch distributed processes: the
            # main_worker process function
            mp.spawn(main_worker, nprocs=ngpus_per_node, args=(ngpus_per_node, args, writer))
        else:
            # Simply call main_worker function
            main_worker(args.gpu, ngpus_per_node, args, writer)


    def main_worker(gpu, ngpus_per_node, args, writer):
        global best_f1
        args.gpu = gpu

        if args.gpu is not None:
            print("Use GPU: {} for training".format(args.gpu))

        if args.distributed:
            if args.dist_url == "env://" and args.rank == -1:
                args.rank = int(os.environ["RANK"])
            if args.multiprocessing_distributed:
                # For multiprocessing distributed training, rank needs to be the
                # global rank among all the processes
                args.rank = args.rank * ngpus_per_node + gpu
            dist.init_process_group(backend=args.dist_backend, init_method=args.dist_url,
                                    world_size=args.world_size, rank=args.rank)
        # create model
        if args.pretrained:
            print("=> using pre-trained model '{}'".format(args.arch))
            model = models.__dict__[args.arch](pretrained=True)
            model.fc = nn.Linear(512 * 1, 4) # replace final classifier
        else:
            print("=> creating model '{}'".format(args.arch))
            if args.arch == 'residual_attention_network':
                from model.residual_attention_network import ResidualAttentionModel_92
                model = ResidualAttentionModel_92(num_classes=4)
            else:
                model = models.__dict__[args.arch](num_classes=4)

        if not torch.cuda.is_available():
            print('using CPU, this will be slow')
        elif args.distributed:
            # For multiprocessing distributed, DistributedDataParallel constructor
            # should always set the single device scope, otherwise,
            # DistributedDataParallel will use all available devices.
            if args.gpu is not None:
                torch.cuda.set_device(args.gpu)
                model.cuda(args.gpu)
                # When using a single GPU per process and per
                # DistributedDataParallel, we need to divide the batch size
                # ourselves based on the total number of GPUs we have
                args.batch_size = int(args.batch_size / ngpus_per_node)
                args.workers = int((args.workers + ngpus_per_node - 1) / ngpus_per_node)
                model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.gpu])
            else:
                model.cuda()
                # DistributedDataParallel will divide and allocate batch_size to all
                # available GPUs if device_ids are not set
                model = torch.nn.parallel.DistributedDataParallel(model)
        elif args.gpu is not None:
            torch.cuda.set_device(args.gpu)
            model = model.cuda(args.gpu)
        else:
            # DataParallel will divide and allocate batch_size to all available GPUs
            if args.arch.startswith('alexnet') or args.arch.startswith('vgg'):
                model.features = torch.nn.DataParallel(model.features)
                model.cuda()
            else:
                model = torch.nn.DataParallel(model).cuda()

        # define loss function (criterion) and optimizer
        criterion = nn.CrossEntropyLoss(reduction='none').cuda(args.gpu)

        # Better Adam optimizer: https://github.com/LiyuanLucasLiu/RAdam
        optimizer = RAdam(model.parameters(),lr=args.lr)

        # optionally resume from a checkpoint
        checkpoint = None
        if args.resume:
            if os.path.isfile(args.resume):
                print("=> loading checkpoint '{}'".format(args.resume))
                if args.gpu is None:
                    checkpoint = torch.load(args.resume, map_location=torch.device('cpu'))
                else:
                    # Map model to be loaded to specified single gpu.
                    loc = 'cuda:{}'.format(args.gpu)
                    checkpoint = torch.load(args.resume, map_location=loc)
                args.start_epoch = checkpoint['epoch']
                best_f1 = checkpoint['best_f1']
                if args.gpu is not None:
                    # best_acc1 may be from a checkpoint from a different GPU
                    best_f1 = best_f1.to(args.gpu)
                model.load_state_dict(checkpoint['state_dict'], strict=False)
                optimizer.load_state_dict(checkpoint['optimizer'])
                print("=> loaded checkpoint '{}' (epoch {})"
                      .format(args.resume, checkpoint['epoch']))
            else:
                print("=> no checkpoint found at '{}'".format(args.resume))

        cudnn.benchmark = True

        # Data loading code
        train_data = args.train_data
        val_data = args.val_data

        if not args.evaluate:
            if not (os.path.isfile(train_data) and os.path.splitext(train_data)[-1] == '.csv'):
                RoofImages.to_csv_datasource(train_data,csv_filename='tmp_train_set.csv', calc_perf=True)
                traindir = 'tmp_train_set.csv'
            else:
                traindir = args.train_data

            if not (os.path.isfile(val_data) and os.path.splitext(val_data)[-1] == '.csv'):
                RoofImages.to_csv_datasource(val_data,csv_filename='tmp_val_set.csv', calc_perf=True)
                valdir = 'tmp_val_set.csv'
            else:
                valdir = args.val_data
        else:
            if not args.resume:
                print('Evaluation is chosen without resuming from a checkpoint. Please choose a checkpoint to load with the --resume parameter.')
                exit(1)

            if not (os.path.isfile(val_data) and os.path.splitext(val_data)[-1] == '.csv'):        
                RoofImages.to_csv_datasource(val_data,csv_filename='tmp_val_set.csv', calc_perf=True)
                valdir = 'tmp_val_set.csv'
            else:
                valdir = args.val_data


        normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                         std=[0.229, 0.224, 0.225])

        if not args.evaluate:
            train_dataset = RoofImages(
                traindir,
                transforms.Compose([
                    transforms.RandomResizedCrop(224),
                    transforms.ColorJitter(0.4,0.4,0.4,0.4),
                    transforms.RandomGrayscale(),
                    transforms.RandomHorizontalFlip(),
                    transforms.RandomVerticalFlip(),
                    transforms.ToTensor(),
                    normalize,
                ]))

            if args.distributed:
                if args.weighted_sampling:
                    print ('Warning: Weighted sampling not implemented for distributed training. So no weighted sampling will be performed')
                train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset)
            else:
                if args.weighted_sampling:
                    train_weights = np.array(train_dataset.train_weights)
                    train_sampler = torch.utils.data.WeightedRandomSampler(train_weights, len(train_weights), replacement=True)
                else:
                    train_sampler = None

            train_loader = torch.utils.data.DataLoader(
                train_dataset, batch_size=args.batch_size, shuffle=(train_sampler is None),
                num_workers=args.workers, pin_memory=True, sampler=train_sampler)

        try:
            classes = checkpoint['classes']
        except:
            classes = None

        val_loader = torch.utils.data.DataLoader(
            RoofImages(valdir, transforms.Compose([
                transforms.Resize(256 if args.val_resize is None else args.val_resize),
                transforms.CenterCrop(224),
                transforms.ToTensor(),
                normalize,
            ]),test_mode=args.evaluate, classes=classes),
            batch_size=args.batch_size, shuffle=False,
            num_workers=args.workers, pin_memory=True)

        if args.evaluate:
            f1, results = validate(val_loader, model, criterion, args, 0, writer, test_mode=True)
            results.to_csv(args.result_file)
            return

        for epoch in range(args.start_epoch, args.epochs):
            if args.distributed:
                train_sampler.set_epoch(epoch)
            adjust_learning_rate(optimizer, epoch, args)

            # train for one epoch
            idx_vec, loss_vec = train(train_loader, model, criterion, optimizer, epoch, args, writer)

            if args.loss_weighting:
                print ('Weighting losses')
                train_set_order = np.argsort(idx_vec)
                loss_vec = (loss_vec - loss_vec.min())/(loss_vec.max() - loss_vec.min())
                loss_vec_in_order = loss_vec[train_set_order]
                train_sampler.weights = torch.as_tensor(loss_vec_in_order, dtype=torch.double)

            # evaluate on validation set
            f1, results = validate(val_loader, model, criterion, args, epoch, writer)
            results.to_csv(args.result_file)

            # remember best acc@1 and save checkpoint
            is_best = f1 > best_f1
            best_f1 = max(f1, best_f1)

            if not args.multiprocessing_distributed or (args.multiprocessing_distributed
                    and args.rank % ngpus_per_node == 0):
                save_checkpoint({
                    'epoch': epoch + 1,
                    'arch': args.arch,
                    'state_dict': model.state_dict(),
                    'best_f1': best_f1,
                    'optimizer' : optimizer.state_dict(),
                    'classes' : train_dataset.classes # Save the exact classnames this checkpoint was created with
                }, is_best=is_best, file_folder = args.log_dir)


    def validate(val_loader, model, criterion, args, epoch, writer, test_mode=False):
        print('Start validation')
        # switch to evaluate mode
        model.eval()
        pred_all = []
        real_all = []
        loss_vec = []
        all_indexes = []

        with torch.no_grad():
            end = time.time()
            for i, (images, target, index) in enumerate(val_loader):
                if args.gpu is not None:
                    images = images.cuda(args.gpu, non_blocking=True)
                if torch.cuda.is_available():
                    target = target.cuda(args.gpu, non_blocking=True)

                # compute output
                output = model(images)

                if not test_mode:

                    loss = criterion(output, target)
                    loss = loss.mean() # Because I need to reduce to none because of something else
                    real_all.append(target.cpu().numpy().flatten())
                    loss_vec.append(loss.item())

                pred = output.max(1, keepdim=True)[1]  # .squeeze() # get the index of the max log-probability
                pred_all.append(pred.flatten().cpu().numpy())

                all_indexes.append(index.flatten().cpu().numpy()) # Making sure the index fits to the label



        if not test_mode:
            y_gt = np.concatenate(real_all)

        y_pred = np.concatenate(pred_all)

        if not test_mode:
            writer.add_scalar('Val/Loss', np.mean(loss_vec), epoch)
            precision, recall, f1, support = precision_recall_fscore_support(y_gt,
                                                                             y_pred, average='macro')
            writer.add_scalar('Val/F1', f1, epoch)
            writer.add_scalar('Val/Precision', precision, epoch)
            writer.add_scalar('Val/Recall', recall, epoch)
            classes = val_loader.dataset.classes


            confusion_matrix = sklearn.metrics.confusion_matrix(y_gt, y_pred, labels=range(len(classes)))
            cm, cm_fig = construct_confusion_matrix_image(classes, confusion_matrix)
            writer.add_image('Val/confusion_matrix', cm, epoch)
        else:
            f1 = None

        all_indexes = np.concatenate(all_indexes)

        filenames = val_loader.dataset.data_df.iloc[all_indexes]['filenames']

        results = pd.DataFrame()
        results['file'] = filenames
    #           
        results['prediction'] = [val_loader.dataset.classes[y_class] for y_class in y_pred]

        return f1, results


    def save_checkpoint(state, is_best, file_folder) :
        filename=os.path.join(file_folder,'checkpoint.pth.tar')
        torch.save(state, filename)
        if is_best:
            shutil.copyfile(filename, os.path.join(file_folder,'model_best.pth.tar'))


    def adjust_learning_rate(optimizer, epoch, args):
        """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
        lr = args.lr * (0.1 ** (epoch // 30))
        for param_group in optimizer.param_groups:
            param_group['lr'] = lr


    def accuracy(output, target, topk=(1,)):
        """Computes the accuracy over the k top predictions for the specified values of k"""
        with torch.no_grad():
            maxk = max(topk)
            batch_size = target.size(0)

            _, pred = output.topk(maxk, 1, True, True)
            pred = pred.t()
            correct = pred.eq(target.view(1, -1).expand_as(pred))

            res = []
            for k in topk:
                correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
                res.append(correct_k.mul_(100.0 / batch_size).cpu().numpy())
            return res



if __name__ == '__main__':
    main()



import pandas as pd
import torch
import torchvision.transforms as transforms
from torch.utils.data import DataLoader
import os
import csv

from .models.resnet_applied import resnet50
from .utils.Datasets import Foundation_Type_Testset

from .csail_segmentation_tool.csail_segmentation import MaskBuilding

class FoundationHeightClassifier():

    def __init__(self, checkpoint='', onlycpu=False, maskBuildings=False,loadMasks=False, workDir='tmp', resultFile='FoundationElevation.csv'):
        '''
        checkpoint (str): Path to checkpoint. Defaults to best pretrained version.
        onlycpu (bool): Use CPU only, disregard GPU by default.
        maskBuildings (bool): Mask the parts of the image which are not a building. Slow.
        loadMasks (bool): Generate a mask on the fly if False and maskBuildings=True
        '''
    
        self.checkpoint = checkpoint
        self.onlycpu = onlycpu
        self.maskBuildings = maskBuildings
        self.loadMasks = loadMasks
        self.workDir = workDir
        self.outFilePath = os.path.join(workDir, resultFile)

        self.checkpointsDir = os.path.join(workDir,'checkpoints')
        os.makedirs(self.checkpointsDir,exist_ok=True)
        weight_file_path = os.path.join(self.checkpointsDir,'best_masked.pkl')
        '''
        if not os.path.isfile(weight_file_path):
            print('Loading remote model file to the weights folder..')
            torch.hub.download_url_to_file('https://zenodo.org/record/4145934/files/best_masked.pkl', weight_file_path)
        '''

        if self.checkpoint != '':
            self.modelFile = self.checkpoint
        else:
            #weight_file_path = os.path.join(self.checkpointsDir,'best_model_weights.pth')
            if not os.path.isfile(weight_file_path):
                print('Loading remote model file to the weights folder..')
                torch.hub.download_url_to_file('https://zenodo.org/record/4145934/files/best_masked.pkl', weight_file_path)
            self.modelFile = weight_file_path

        # need to change this to tmp folder
        model_name='ade20k-resnet50dilated-ppm_deepsup'
        model_dir=os.path.join(workDir, 'csail_segmentation_tool','csail_seg',model_name)
        os.makedirs(model_dir, exist_ok=True)

        encoder=f'{model_name}/encoder_epoch_20.pth'
        decoder=f'{model_name}/decoder_epoch_20.pth'
        localEncoderFilePath = os.path.join(model_dir,'encoder_epoch_20.pth')
        localDecoderFilePath = os.path.join(model_dir,'decoder_epoch_20.pth')
        if not os.path.isfile(localEncoderFilePath):
            print('Loading remote model (encoder) file to the weights folder..')
            torch.hub.download_url_to_file(f'http://sceneparsing.csail.mit.edu/model/pytorch/{encoder}',localEncoderFilePath)

        if not os.path.isfile(localDecoderFilePath):
            print('Loading remote model (decoder) file to the weights folder..')
            torch.hub.download_url_to_file(f'http://sceneparsing.csail.mit.edu/model/pytorch/{decoder}',localDecoderFilePath)

        if self.onlycpu: 
            self.device='cpu'
        else: 
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'


        # test_transforms
        normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                         std=[0.229, 0.224, 0.225])

        self.test_transforms = [transforms.Resize((224, 224)),
                          transforms.ToTensor(),
                          normalize]

        if self.maskBuildings and not self.loadMasks:
            #from csail_segmentation_tool.csail_segmentation import MaskBuilding
            self.test_transforms.insert(0, MaskBuilding(self.device, model_dir=model_dir))

        self.test_transforms = transforms.Compose(self.test_transforms)

    def predict(self,image=''):
        '''
        image (str): Path to one image or a folder containing images.
        '''

        dataset = Foundation_Type_Testset(image, transform=self.test_transforms, mask_buildings=self.maskBuildings,
                                          load_masks = self.loadMasks)

        test_loader = DataLoader(dataset, batch_size=1, shuffle=False, num_workers=0)
        model = resnet50(low_dim=1)


        model_file = self.modelFile

        if not torch.cuda.is_available():
            state_dict = torch.load(model_file, map_location=torch.device('cpu'))
        else:
            state_dict = torch.load(model_file)


        missing, unexpected = model.load_state_dict(state_dict, strict=False)
        if any(['module' in name for name in unexpected]):
            # Remapping to remove effects of DataParallel
            from collections import OrderedDict
            new_state_dict = OrderedDict()
            for k, v in state_dict.items():
                name = k[7:]  # remove 'module.' of dataparallel
                new_state_dict[name] = v

            model.load_state_dict(new_state_dict, strict=False)
        else:
            if len(missing) or len(unexpected):
                print('Missing or unexpected keys: {},{}'.format(missing, unexpected))
                print('This should not happen. Check if checkpoint is correct')

        model.eval()
        model = model.to(self.device)

        predictions = []
        probs = []
        imagePathList = []

        with torch.no_grad():
            for i, (images, filename) in enumerate(test_loader):
                images = images.to(self.device)

                prediction = model(images.float())
                score = torch.sigmoid(prediction).cpu().numpy()[0][0]

                p = int( score >= 0.5) # class: 0 or 1
                predictions.append(p)
                imagePathList.append(filename[0])
                prob = score if score >= 0.5 else 1.-score
                probs.append(prob)
                print(f"Image :  {filename[0]}     Class : {p} ({str(round(prob*100,2))}%)") 

        
        df = pd.DataFrame(list(zip(imagePathList, predictions, probs)), columns =['image', 'prediction', 'probability']) 
        df.to_csv(self.outFilePath, index=False)
        print(f'Results written in file {self.outFilePath}')

        return df


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
/*------------------------------------------------------*
|                         BRAILS                        |
|                                                       |
| Author: Charles Wang,  UC Berkeley, c_w@berkeley.edu  |
|                                                       |
| Date:    10/15/2020                                   |
*------------------------------------------------------*/
"""

import matplotlib.pyplot as plt
import itertools
import numpy as np

def plot_confusion_matrix(cm, classes,
                          normalize=False,
                          title='Confusion matrix',
                          cmap=plt.cm.Blues,
                          xlabel='True label',ylabel='Predicted label'):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        print("Normalized confusion matrix")
    else:
        print('Confusion matrix, without normalization')
    
    fig, ax = plt.subplots()
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=0)
    plt.yticks(tick_marks, classes)


    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, round(cm[i, j],1),
                 horizontalalignment="center",
                 color="white" if cm[i, j] > thresh else "black")

    plt.tight_layout()
    plt.ylabel(ylabel, fontsize=12)
    plt.xlabel(xlabel, fontsize=12)
    plt.xticks(fontsize=10)
    plt.yticks(fontsize=10)
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")
    plt.show()



import torch.nn as nn
import math
import torch.utils.model_zoo as model_zoo

__all__ = ['ResNet', 'resnet18', 'resnet34', 'resnet50', 'resnet101',
           'resnet152']

model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
}

def conv3x3(in_planes, out_planes, stride=1):
    "3x3 convolution with padding"
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self, block, layers, low_dim=128):
        self.inplanes = 64
        super(ResNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)
        self.avgpool = nn.AvgPool2d(7, stride=1)
        self.fc = nn.Linear(512 * block.expansion, low_dim)
        #self.l2norm = Normalize(2)
        # When applied the normalize is optional and done in later steps

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)


        return x


def resnet18(pretrained=False, **kwargs):
    """Constructs a ResNet-18 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['resnet18']))
    return model


def resnet34(pretrained=False, **kwargs):
    """Constructs a ResNet-34 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['resnet34']))
    return model


def resnet50(pretrained=False, **kwargs):
    """Constructs a ResNet-50 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    if 'finetune' in kwargs.keys() and not kwargs['finetune']:
        model = ResNet(Bottleneck, [3, 4, 6, 3], low_dim=kwargs['low_dim'])
    elif pretrained and 'finetune' in kwargs.keys() and kwargs['finetune']:
        model = ResNet(Bottleneck, [3, 4, 6, 3], low_dim=1000)
        model.load_state_dict(model_zoo.load_url(model_urls['resnet50']))
        num_ftrs = model.fc.in_features
        model.fc = nn.Linear(num_ftrs, kwargs['low_dim'])

    elif pretrained:
        model = ResNet(Bottleneck, [3, 4, 6, 3], low_dim=kwargs['low_dim'])
        model.load_state_dict(model_zoo.load_url(model_urls['resnet50']))
    else:
        model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    return model


def resnet101(pretrained=False, **kwargs):
    """Constructs a ResNet-101 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['resnet101']))
    return model


def resnet152(pretrained=False, **kwargs):
    """Constructs a ResNet-152 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['resnet152']))
    return model



import torch.nn as nn
from  torch.utils.model_zoo import load_url as load_state_dict_from_url


__all__ = ['ResNet', 'resnet18', 'resnet34', 'resnet50', 'resnet101',
           'resnet152', 'resnext50_32x4d', 'resnext101_32x8d']


model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
}


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNet(nn.Module):

    def __init__(self, block, layers, num_classes=1000, zero_init_residual=False,
                 groups=1, width_per_group=64, replace_stride_with_dilation=None,
                 norm_layer=None):
        super(ResNet, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer

        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError("replace_stride_with_dilation should be None "
                             "or a 3-element tuple, got {}".format(replace_stride_with_dilation))
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2,
                                       dilate=replace_stride_with_dilation[0])
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2,
                                       dilate=replace_stride_with_dilation[1])
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2,
                                       dilate=replace_stride_with_dilation[2])
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x1 = self.maxpool(x)

        x2 = self.layer1(x1)
        x3 = self.layer2(x2)
        x4 = self.layer3(x3)
        x5 = self.layer4(x4)

        x = self.avgpool(x5)
        x = x.reshape(x.size(0), -1)
        x = self.fc(x)

        return x1, x2, x3, x4, x5


def _resnet(arch, inplanes, planes, pretrained, progress, **kwargs):
    model = ResNet(inplanes, planes, **kwargs)
    if pretrained:
        state_dict = load_state_dict_from_url(model_urls[arch],
                                              progress=progress)
        model.load_state_dict(state_dict)
    return model


def resnet18(pretrained=False, progress=True, **kwargs):
    """Constructs a ResNet-18 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet18', BasicBlock, [2, 2, 2, 2], pretrained, progress,
                   **kwargs)


def resnet34(pretrained=False, progress=True, **kwargs):
    """Constructs a ResNet-34 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet34', BasicBlock, [3, 4, 6, 3], pretrained, progress,
                   **kwargs)


def resnet50(pretrained=False, progress=True, **kwargs):
    """Constructs a ResNet-50 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet50', Bottleneck, [3, 4, 6, 3], pretrained, progress,
                   **kwargs)


def resnet101(pretrained=False, progress=True, **kwargs):
    """Constructs a ResNet-101 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet101', Bottleneck, [3, 4, 23, 3], pretrained, progress,
                   **kwargs)


def resnet152(pretrained=False, progress=True, **kwargs):
    """Constructs a ResNet-152 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnet('resnet152', Bottleneck, [3, 8, 36, 3], pretrained, progress,
                   **kwargs)


def resnext50_32x4d(**kwargs):
    kwargs['groups'] = 32
    kwargs['width_per_group'] = 4
    return _resnet('resnext50_32x4d', Bottleneck, [3, 4, 6, 3],
                   pretrained=False, progress=True, **kwargs)


def resnext101_32x8d(**kwargs):
    kwargs['groups'] = 32
    kwargs['width_per_group'] = 8
    return _resnet('resnext101_32x8d', Bottleneck, [3, 4, 23, 3],
                   pretrained=False, progress=True, **kwargs)



# Author: Zylo117

import os

import cv2
import numpy as np
import torch
from glob import glob
from torch import nn
from torchvision.ops import nms
from torchvision.ops.boxes import batched_nms
from typing import Union
import uuid

from lib.utils.sync_batchnorm import SynchronizedBatchNorm2d

from torch.nn.init import _calculate_fan_in_and_fan_out, _no_grad_normal_
import math
import webcolors

def invert_affine(metas: Union[float, list, tuple], preds):
    for i in range(len(preds)):
        if len(preds[i]['rois']) == 0:
            continue
        else:
            if metas is float:
                preds[i]['rois'][:, [0, 2]] = preds[i]['rois'][:, [0, 2]] / metas
                preds[i]['rois'][:, [1, 3]] = preds[i]['rois'][:, [1, 3]] / metas
            else:
                new_w, new_h, old_w, old_h, padding_w, padding_h = metas[i]
                preds[i]['rois'][:, [0, 2]] = preds[i]['rois'][:, [0, 2]] / (new_w / old_w)
                preds[i]['rois'][:, [1, 3]] = preds[i]['rois'][:, [1, 3]] / (new_h / old_h)
    return preds


def aspectaware_resize_padding(image, width, height, interpolation=None, means=None):
    old_h, old_w, c = image.shape
    if old_w > old_h:
        new_w = width
        new_h = int(width / old_w * old_h)
    else:
        new_w = int(height / old_h * old_w)
        new_h = height

    canvas = np.zeros((height, height, c), np.float32)
    if means is not None:
        canvas[...] = means

    if new_w != old_w or new_h != old_h:
        if interpolation is None:
            image = cv2.resize(image, (new_w, new_h))
        else:
            image = cv2.resize(image, (new_w, new_h), interpolation=interpolation)

    padding_h = height - new_h
    padding_w = width - new_w

    if c > 1:
        canvas[:new_h, :new_w] = image
    else:
        if len(image.shape) == 2:
            canvas[:new_h, :new_w, 0] = image
        else:
            canvas[:new_h, :new_w] = image

    return canvas, new_w, new_h, old_w, old_h, padding_w, padding_h,


def preprocess(*image_path, max_size=512, mean=(0.406, 0.456, 0.485), std=(0.225, 0.224, 0.229)):
    ori_imgs = [cv2.imread(img_path) for img_path in image_path]
    normalized_imgs = [(img / 255 - mean) / std for img in ori_imgs]
    imgs_meta = [aspectaware_resize_padding(img[..., ::-1], max_size, max_size,
                                            means=None) for img in normalized_imgs]
    framed_imgs = [img_meta[0] for img_meta in imgs_meta]
    framed_metas = [img_meta[1:] for img_meta in imgs_meta]

    return ori_imgs, framed_imgs, framed_metas


def preprocess_video(*frame_from_video, max_size=512, mean=(0.406, 0.456, 0.485), std=(0.225, 0.224, 0.229)):
    ori_imgs = frame_from_video
    normalized_imgs = [(img / 255 - mean) / std for img in ori_imgs]
    imgs_meta = [aspectaware_resize_padding(img[..., ::-1], max_size, max_size,
                                            means=None) for img in normalized_imgs]
    framed_imgs = [img_meta[0] for img_meta in imgs_meta]
    framed_metas = [img_meta[1:] for img_meta in imgs_meta]

    return ori_imgs, framed_imgs, framed_metas


def postprocess(x, anchors, regression, classification, regressBoxes, clipBoxes, threshold, iou_threshold):
    transformed_anchors = regressBoxes(anchors, regression)
    transformed_anchors = clipBoxes(transformed_anchors, x)
    scores = torch.max(classification, dim=2, keepdim=True)[0]
    scores_over_thresh = (scores > threshold)[:, :, 0]
    out = []
    for i in range(x.shape[0]):
        if scores_over_thresh[i].sum() == 0:
            out.append({
                'rois': np.array(()),
                'class_ids': np.array(()),
                'scores': np.array(()),
            })
            continue

        classification_per = classification[i, scores_over_thresh[i, :], ...].permute(1, 0)
        transformed_anchors_per = transformed_anchors[i, scores_over_thresh[i, :], ...]
        scores_per = scores[i, scores_over_thresh[i, :], ...]
        scores_, classes_ = classification_per.max(dim=0)
        anchors_nms_idx = batched_nms(transformed_anchors_per, scores_per[:, 0], classes_, iou_threshold=iou_threshold)

        if anchors_nms_idx.shape[0] != 0:
            classes_ = classes_[anchors_nms_idx]
            scores_ = scores_[anchors_nms_idx]
            boxes_ = transformed_anchors_per[anchors_nms_idx, :]

            out.append({
                'rois': boxes_.cpu().numpy(),
                'class_ids': classes_.cpu().numpy(),
                'scores': scores_.cpu().numpy(),
            })
        else:
            out.append({
                'rois': np.array(()),
                'class_ids': np.array(()),
                'scores': np.array(()),
            })

    return out


def display(preds, imgs, obj_list, imshow=True, imwrite=False):
    for i in range(len(imgs)):
        if len(preds[i]['rois']) == 0:
            continue

        for j in range(len(preds[i]['rois'])):
            (x1, y1, x2, y2) = preds[i]['rois'][j].astype(np.int)
            obj = obj_list[preds[i]['class_ids'][j]]
            score = float(preds[i]['scores'][j])

            plot_one_box(imgs[i], [x1, y1, x2, y2], label=obj, score=score,
                         color=color_list[get_index_label(obj, obj_list)])
        if imshow:
            cv2.imshow('img', imgs[i])
            cv2.waitKey(0)

        if imwrite:
            os.makedirs('test/', exist_ok=True)
            cv2.imwrite(f'test/{uuid.uuid4().hex}.jpg', imgs[i])


def replace_w_sync_bn(m):
    for var_name in dir(m):
        target_attr = getattr(m, var_name)
        if type(target_attr) == torch.nn.BatchNorm2d:
            num_features = target_attr.num_features
            eps = target_attr.eps
            momentum = target_attr.momentum
            affine = target_attr.affine

            # get parameters
            running_mean = target_attr.running_mean
            running_var = target_attr.running_var
            if affine:
                weight = target_attr.weight
                bias = target_attr.bias

            setattr(m, var_name,
                    SynchronizedBatchNorm2d(num_features, eps, momentum, affine))

            target_attr = getattr(m, var_name)
            # set parameters
            target_attr.running_mean = running_mean
            target_attr.running_var = running_var
            if affine:
                target_attr.weight = weight
                target_attr.bias = bias

    for var_name, children in m.named_children():
        replace_w_sync_bn(children)


class CustomDataParallel(nn.DataParallel):
    """
    force splitting data to all gpus instead of sending all data to cuda:0 and then moving around.
    """

    def __init__(self, module, num_gpus):
        super().__init__(module)
        self.num_gpus = num_gpus

    def scatter(self, inputs, kwargs, device_ids):
        # More like scatter and data prep at the same time. The point is we prep the data in such a way
        # that no scatter is necessary, and there's no need to shuffle stuff around different GPUs.
        devices = ['cuda:' + str(x) for x in range(self.num_gpus)]
        splits = inputs[0].shape[0] // self.num_gpus

        if splits == 0:
            raise Exception('Batchsize must be greater than num_gpus.')

        return [(inputs[0][splits * device_idx: splits * (device_idx + 1)].to(f'cuda:{device_idx}', non_blocking=True),
                 inputs[1][splits * device_idx: splits * (device_idx + 1)].to(f'cuda:{device_idx}', non_blocking=True))
                for device_idx in range(len(devices))], \
               [kwargs] * len(devices)


def get_last_weights(weights_path):
    weights_path = glob(weights_path + f'/*.pth')
    weights_path = sorted(weights_path,
                          key=lambda x: int(x.rsplit('_')[-1].rsplit('.')[0]),
                          reverse=True)[0]
    print(f'using weights {weights_path}')
    return weights_path


def init_weights(model):
    for name, module in model.named_modules():
        is_conv_layer = isinstance(module, nn.Conv2d)

        if is_conv_layer:
            if "conv_list" or "header" in name:
                variance_scaling_(module.weight.data)
            else:
                nn.init.kaiming_uniform_(module.weight.data)

            if module.bias is not None:
                if "classifier.header" in name:
                    bias_value = -np.log((1 - 0.01) / 0.01)
                    torch.nn.init.constant_(module.bias, bias_value)
                else:
                    module.bias.data.zero_()


def variance_scaling_(tensor, gain=1.):
    # type: (Tensor, float) -> Tensor
    r"""
    initializer for SeparableConv in Regressor/Classifier
    reference: https://keras.io/zh/initializers/  VarianceScaling
    """
    fan_in, fan_out = _calculate_fan_in_and_fan_out(tensor)
    std = math.sqrt(gain / float(fan_in))

    return _no_grad_normal_(tensor, 0., std)


STANDARD_COLORS = [
    'LawnGreen', 'Chartreuse', 'Aqua','Beige', 'Azure','BlanchedAlmond','Bisque',
    'Aquamarine', 'BlueViolet', 'BurlyWood', 'CadetBlue', 'AntiqueWhite',
    'Chocolate', 'Coral', 'CornflowerBlue', 'Cornsilk', 'Crimson', 'Cyan',
    'DarkCyan', 'DarkGoldenRod', 'DarkGrey', 'DarkKhaki', 'DarkOrange',
    'DarkOrchid', 'DarkSalmon', 'DarkSeaGreen', 'DarkTurquoise', 'DarkViolet',
    'DeepPink', 'DeepSkyBlue', 'DodgerBlue', 'FireBrick', 'FloralWhite',
    'ForestGreen', 'Fuchsia', 'Gainsboro', 'GhostWhite', 'Gold', 'GoldenRod',
    'Salmon', 'Tan', 'HoneyDew', 'HotPink', 'IndianRed', 'Ivory', 'Khaki',
    'Lavender', 'LavenderBlush', 'AliceBlue', 'LemonChiffon', 'LightBlue',
    'LightCoral', 'LightCyan', 'LightGoldenRodYellow', 'LightGray', 'LightGrey',
    'LightGreen', 'LightPink', 'LightSalmon', 'LightSeaGreen', 'LightSkyBlue',
    'LightSlateGray', 'LightSlateGrey', 'LightSteelBlue', 'LightYellow', 'Lime',
    'LimeGreen', 'Linen', 'Magenta', 'MediumAquaMarine', 'MediumOrchid',
    'MediumPurple', 'MediumSeaGreen', 'MediumSlateBlue', 'MediumSpringGreen',
    'MediumTurquoise', 'MediumVioletRed', 'MintCream', 'MistyRose', 'Moccasin',
    'NavajoWhite', 'OldLace', 'Olive', 'OliveDrab', 'Orange', 'OrangeRed',
    'Orchid', 'PaleGoldenRod', 'PaleGreen', 'PaleTurquoise', 'PaleVioletRed',
    'PapayaWhip', 'PeachPuff', 'Peru', 'Pink', 'Plum', 'PowderBlue', 'Purple',
    'Red', 'RosyBrown', 'RoyalBlue', 'SaddleBrown', 'Green', 'SandyBrown',
    'SeaGreen', 'SeaShell', 'Sienna', 'Silver', 'SkyBlue', 'SlateBlue',
    'SlateGray', 'SlateGrey', 'Snow', 'SpringGreen', 'SteelBlue', 'GreenYellow',
    'Teal', 'Thistle', 'Tomato', 'Turquoise', 'Violet', 'Wheat', 'White',
    'WhiteSmoke', 'Yellow', 'YellowGreen'
]


def from_colorname_to_bgr(color):
    rgb_color=webcolors.name_to_rgb(color)
    result=(rgb_color.blue,rgb_color.green,rgb_color.red)
    return result


def standard_to_bgr(list_color_name):
    standard= []
    for i in range(len(list_color_name)-36): #-36 used to match the len(obj_list)
        standard.append(from_colorname_to_bgr(list_color_name[i]))
    return standard


def get_index_label(label, obj_list):
    index = int(obj_list.index(label))
    return index


def plot_one_box(img, coord, label=None, score=None, color=None, line_thickness=None):
    tl = line_thickness or int(round(0.001 * max(img.shape[0:2])))  # line thickness
    color = color
    c1, c2 = (int(coord[0]), int(coord[1])), (int(coord[2]), int(coord[3]))
    cv2.rectangle(img, c1, c2, color, thickness=tl)
    if label:
        tf = max(tl - 2, 1)  # font thickness
        s_size = cv2.getTextSize(str('{:.0%}'.format(score)),0, fontScale=float(tl) / 3, thickness=tf)[0]
        t_size = cv2.getTextSize(label, 0, fontScale=float(tl) / 3, thickness=tf)[0]
        c2 = c1[0] + t_size[0]+s_size[0]+15, c1[1] - t_size[1] -3
        cv2.rectangle(img, c1, c2 , color, -1)  # filled
        cv2.putText(img, '{}: {:.0%}'.format(label, score), (c1[0],c1[1] - 2), 0, float(tl) / 3, [0, 0, 0], thickness=tf, lineType=cv2.FONT_HERSHEY_SIMPLEX)

        
color_list = standard_to_bgr(STANDARD_COLORS)



import torch
import time

from lib.utils import AverageMeter
import torchvision.transforms as transforms
import numpy as np

def NN(epoch, net, lemniscate, trainloader, testloader, recompute_memory=0):
    net.eval()
    net_time = AverageMeter()
    cls_time = AverageMeter()
    losses = AverageMeter()
    correct = 0.
    total = 0
    testsize = testloader.dataset.__len__()

    trainFeatures = lemniscate.memory.t()
    if hasattr(trainloader.dataset, 'imgs'):
        trainLabels = torch.LongTensor([y for y in trainloader.dataset.targets]).cuda()
    else:
        trainLabels = torch.ShortTensor(trainloader.dataset.train_labels).cuda()

    if recompute_memory:
        transform_bak = trainloader.dataset.transform
        trainloader.dataset.transform = testloader.dataset.transform
        temploader = torch.utils.data.DataLoader(trainloader.dataset, batch_size=100, shuffle=False, num_workers=1)
        for batch_idx, (inputs, targets, indexes) in enumerate(temploader):
            targets = targets.cuda(non_blocking=True)
            batchSize = inputs.size(0)
            features = net(inputs)
            trainFeatures[:, batch_idx*batchSize:batch_idx*batchSize+batchSize] = features.data.t()
        trainLabels = torch.LongTensor(temploader.dataset.train_labels).cuda()
        trainloader.dataset.transform = transform_bak
    
    end = time.time()
    with torch.no_grad():
        for batch_idx, (inputs, targets, indexes) in enumerate(testloader):
            targets = targets.cuda(non_blocking=True)
            batchSize = inputs.size(0)
            features = net(inputs)
            net_time.update(time.time() - end)
            end = time.time()

            dist = torch.mm(features, trainFeatures)

            yd, yi = dist.topk(1, dim=1, largest=True, sorted=True)
            candidates = trainLabels.view(1,-1).expand(batchSize, -1)
            retrieval = torch.gather(candidates, 1, yi)

            retrieval = retrieval.narrow(1, 0, 1).clone().view(-1)
            yd = yd.narrow(1, 0, 1)

            total += targets.size(0)
            correct += retrieval.eq(targets.data).sum().item()
            
            cls_time.update(time.time() - end)
            end = time.time()

            print('Test [{}/{}]\t'
                  'Net Time {net_time.val:.3f} ({net_time.avg:.3f})\t'
                  'Cls Time {cls_time.val:.3f} ({cls_time.avg:.3f})\t'
                  'Top1: {:.2f}'.format(
                  total, testsize, correct*100./total, net_time=net_time, cls_time=cls_time))

    return correct/total

def kNN(epoch, net, lemniscate, trainloader, testloader, K, sigma, recompute_memory=0):
    net.eval()
    net_time = AverageMeter()
    cls_time = AverageMeter()
    total = 0
    testsize = testloader.dataset.__len__()

    trainFeatures = lemniscate.memory.t()
    if hasattr(trainloader.dataset, 'imgs'):
        trainLabels = torch.LongTensor([y for (p, y) in trainloader.dataset.imgs]).cuda()
    else:
        trainLabels = torch.LongTensor(trainloader.dataset.train_labels).cuda()
    C = trainLabels.max() + 1

    if recompute_memory:
        transform_bak = trainloader.dataset.transform
        trainloader.dataset.transform = testloader.dataset.transform
        temploader = torch.utils.data.DataLoader(trainloader.dataset, batch_size=100, shuffle=False, num_workers=1)
        for batch_idx, (inputs, targets, indexes) in enumerate(temploader):
            targets = targets.cuda(non_blocking=True)
            batchSize = inputs.size(0)
            features = net(inputs)
            trainFeatures[:, batch_idx*batchSize:batch_idx*batchSize+batchSize] = features.data.t()
        trainLabels = torch.LongTensor(temploader.dataset.train_labels).cuda()
        trainloader.dataset.transform = transform_bak
    
    top1 = 0.
    top5 = 0.
    end = time.time()
    with torch.no_grad():
        retrieval_one_hot = torch.zeros(K, C).cuda()
        for batch_idx, (inputs, targets, indexes) in enumerate(testloader):
            end = time.time()
            targets = targets.cuda(non_blocking=True)
            batchSize = inputs.size(0)
            features = net(inputs)
            net_time.update(time.time() - end)
            end = time.time()

            dist = torch.mm(features, trainFeatures)

            yd, yi = dist.topk(K, dim=1, largest=True, sorted=True)
            candidates = trainLabels.view(1,-1).expand(batchSize, -1)
            retrieval = torch.gather(candidates, 1, yi)

            retrieval_one_hot.resize_(batchSize * K, C).zero_()
            retrieval_one_hot.scatter_(1, retrieval.view(-1, 1), 1)
            yd_transform = yd.clone().div_(sigma).exp_()
            probs = torch.sum(torch.mul(retrieval_one_hot.view(batchSize, -1 , C), yd_transform.view(batchSize, -1, 1)), 1)
            _, predictions = probs.sort(1, True)

            # Find which predictions match the target
            correct = predictions.eq(targets.data.view(-1,1))
            cls_time.update(time.time() - end)

            top1 = top1 + correct.narrow(1,0,1).sum().item()
            top5 = top5 + correct.narrow(1,0,5).sum().item()

            total += targets.size(0)

            print('Test [{}/{}]\t'
                  'Net Time {net_time.val:.3f} ({net_time.avg:.3f})\t'
                  'Cls Time {cls_time.val:.3f} ({cls_time.avg:.3f})\t'
                  'Top1: {:.2f}  Top5: {:.2f}'.format(
                  total, testsize, top1*100./total, top5*100./total, net_time=net_time, cls_time=cls_time))

    print(top1*100./total)

    return top1/total




# -*- coding: utf-8 -*-
"""
/*------------------------------------------------------*
|                         BRAILS                        |
|                                                       |
| Author: Charles Wang,  UC Berkeley, c_w@berkeley.edu  |
|                                                       |
| Date:    02/13/2021                                   |
*------------------------------------------------------*/
"""

from brails.modules.ModelZoo import zoo
import wget 
import os
from brails.modules.SoftstoryDetector.GenericDetector import *
import pandas as pd

class SoftstoryDetector(GenericDetector):
    """ Softstory Detector. """


    def __init__(self, modelName=None, classNames=None, resultFile='softstory2_preds.csv', workDir='tmp', printRes=True):
        '''
        modelFile: path to the model
        classNames: a list of classnames
        '''

        if not os.path.exists(workDir): os.makedirs(workDir)

        fileURL = zoo['softstory']['fileURL']
        
        if not classNames:
            classNames = zoo['softstory']['classNames']

        if not modelName:
            modelName = 'softstory_seg_v1.0'
            print('A default softstory model will be used: {}.'.format(modelName))

        modelFile = os.path.join(workDir,'{}.pth'.format(modelName))


        if not os.path.exists(modelFile): # download
            print('Downloading the model ...')
            downloadedModelFile = wget.download(fileURL, out=modelFile)

        GenericDetector.__init__(self, modelName=modelName, classNames=classNames, resultFile=resultFile, printRes=printRes)

    def detectOne(self, imagePath):

        img = cv2.imread(str(imagePath))
        outputs = self.predictor(img)
        instances = outputs["instances"].to("cpu")   

        # area:   instances[0].pred_masks.sum()
        if len(instances)>0:
            preds = instances.pred_classes.tolist() # 0: others ; 1:softstory
            scores = instances.scores.tolist()
            areas = [int(instances.pred_masks[i].sum()) for i in range(len(instances))]
            i = areas.index(max(areas))
            return preds[i], scores[i]

        return None, None

    def detect(self, images):
        preds, scores = [], []
        if isinstance(images, str):
            p, s = self.detectOne(images)
            preds.append(p)
            scores.append(s)
        elif isinstance(images, list):
            for img in images:
                p, s = self.detectOne(img)
                preds.append(p)
                scores.append(s)
                
        else: pass

        df = pd.DataFrame(list(zip(images, preds, scores)), columns =['image', 'prediction', 'probability']) 
        df.to_csv(self.resultFile, index=False)
        print('Results written in file {}'.format(self.resultFile))

        return df


        

if __name__ == '__main__':
    main()









# -*- coding: utf-8 -*-
"""
/*------------------------------------------------------*
|                         BRAILS                        |
|                                                       |
| Author: Charles Wang,  UC Berkeley, c_w@berkeley.edu  |
|                                                       |
| Date:    10/15/2020                                   |
*------------------------------------------------------*/
"""

import argparse
import os
import random
import tensorflow as tf
import numpy as np
from glob import glob
import pandas as pd
from tensorflow.keras.preprocessing import image
from tensorflow.keras.models import load_model
from tensorflow.keras.preprocessing import image_dataset_from_directory
from tensorflow.keras.applications.inception_v3 import preprocess_input
from brails.modules.ModelZoo import zoo
import matplotlib.pyplot as plt
import json

from brails.utils.plotUtils import plot_confusion_matrix
from sklearn.metrics import confusion_matrix
from sklearn.metrics import f1_score, accuracy_score

class ImageClassifier:
    """ A Generic Image Classifier. """

    def __init__(self, modelName=None, classNames=None, resultFile='preds.csv', workDir='tmp', printRes=True):
        '''
        modelFile: path to the model
        classNames: a list of classnames
        '''

        if not os.path.exists(workDir): os.makedirs(workDir)

        if not modelName:

            modelName = 'myCoolModelv0.1'
            print('You didn\'t specify modelName, a default one is assigned {}.'.format(modelName))

        modelFile = os.path.join(workDir,'{}.h5'.format(modelName))
        modelDetailFile = os.path.join(workDir,'{}.json'.format(modelName))

        self.workDir = workDir
        self.modelFile = modelFile
        self.classNames = classNames
        self.resultFile = resultFile
        self.modelName = modelName
        self.modelDetailFile = modelDetailFile
        self.printRes = printRes

        if os.path.exists(modelFile):
            self.model = load_model(modelFile)
            print('Model found locally: {} '.format(modelFile))
            
            # check if a local definition of the model exists.
            if os.path.exists(self.modelDetailFile):
                with open(self.modelDetailFile) as f:
                    self.classNames = json.load(f)['classNames']
                    print('Class names found in the detail file: {} '.format(self.classNames))

        else:
            print('Model file {} doesn\'t exist locally. You are going to train your own model.'.format(modelFile))


    def predictOne(self,imagePath):

        img = image.load_img(imagePath, target_size=(256, 256))
        x = image.img_to_array(img)
        x = np.expand_dims(x, axis=0)
        prediction = self.model.predict(x)
        prob = max(prediction[0])
        prediction = np.argmax(prediction[0])
        if self.classNames: prediction = self.classNames[prediction]

        if os.path.getsize(imagePath)/1024 < 9: # small image, likely to be empty
            #print("{imagePath} is blank. No predictions.")
            #return [imagePath,None, None]
            print("Image :  {}     Class : {} ({}%)".format(imagePath, prediction, str(round(0*100,2)))) 
            return [imagePath,prediction,0]
        else:
            print("Image :  {}     Class : {} ({}%)".format(imagePath, prediction, str(round(prob*100,2)))) 
            return [imagePath,prediction,prob]

    def predictMulti(self,imagePathList):
        predictions = []
        probs = []
        for imagePath in imagePathList:
            '''
            if os.path.getsize(imagePath)/1024 < 9: # small image, likely to be empty
                probs.append(0)
                predictions.append(None)
            else:
                img = image.load_img(imagePath, target_size=(256, 256))
                x = image.img_to_array(img)
                x = np.expand_dims(x, axis=0)
                prediction = self.model.predict(x)
                probs.append(max(prediction[0]))
                prediction = np.argmax(prediction[0])
                if self.classNames: prediction = self.classNames[prediction]
                predictions.append(prediction)
            '''
            img = image.load_img(imagePath, target_size=(256, 256))
            x = image.img_to_array(img)
            x = np.expand_dims(x, axis=0)
            prediction = self.model.predict(x)
            if os.path.getsize(imagePath)/1024 < 9: # small image, likely to be empty
                probs.append(0)
            else:
                probs.append(max(prediction[0]))
            prediction = np.argmax(prediction[0])
            if self.classNames: prediction = self.classNames[prediction]
            predictions.append(prediction)

        if self.printRes:
            for img, pred, prob in zip(imagePathList, predictions, probs): 
                print("Image :  {}     Class : {} ({}%)".format(img, pred, str(round(prob*100,2)))) 

        df = pd.DataFrame(list(zip(imagePathList, predictions, probs)), columns =['image', 'prediction', 'probability']) 
        df.to_csv(self.resultFile, index=False)
        print('Results written in file {}'.format(self.resultFile))

        return df
    
    def predict(self,image):
        if type(image) is list: pred = self.predictMulti(image)
        elif type(image) is str: pred = self.predictOne(image)
        else: 
            print("The parameter of this function should be string or list.")
            pred = []
        return pred

    def loadData(self, imgDir, randomseed=1993, image_size=(256, 256), batch_size = 32, split=[0.8,0.2]):

        #print('* First split the data with 8:2.')
        self.train_ds = image_dataset_from_directory(imgDir,
        validation_split=split[1],
        subset="training",
        seed=randomseed,
        image_size=image_size,
        batch_size=batch_size,
        label_mode='categorical')

        
        self.val_ds = image_dataset_from_directory(
            imgDir,
            validation_split=split[1],
            subset="validation",
            seed=randomseed,
            image_size=image_size,
            batch_size=batch_size,
            label_mode='categorical'
        )
        

        self.image_size = image_size
        self.batch_size = batch_size

        newClassNames = self.train_ds.class_names

        # Configure the dataset for performance 
        AUTOTUNE = tf.data.experimental.AUTOTUNE
        self.train_ds = self.train_ds.prefetch(buffer_size=AUTOTUNE)
        self.val_ds = self.val_ds.prefetch(buffer_size=AUTOTUNE)

        if self.classNames == None:
            self.classNames = newClassNames
        print('The names of the classes are: ', newClassNames)

        if newClassNames != self.classNames:
            print('Error. Folder names {} mismatch predefined classNames: {}'.format(newClassNames, self.classNames))
            return

    def train(self,baseModel='InceptionV3',lr1=0.0001,initial_epochs=10,
                    fine_tune_at=300,lr2=0.00001,fine_tune_epochs=50,
                    horizontalFlip=False,verticalFlip=False,dropout=0.6,randomRotation=0.0,plot=True):
        
        ## 1. Model zoo
         
        modelDict = {
            'Xception': tf.keras.applications.Xception	,
            'VGG16': tf.keras.applications.VGG16	    ,
            'VGG19': tf.keras.applications.VGG19	    ,
            'ResNet50': tf.keras.applications.ResNet50	, # 175
            'ResNet101': tf.keras.applications.ResNet101	, #345
            'ResNet152': tf.keras.applications.ResNet152	,
            'ResNet50V2': tf.keras.applications.ResNet50V2	,
            'ResNet101V2': tf.keras.applications.ResNet101V2,	
            'ResNet152V2': tf.keras.applications.ResNet152V2,	
            'InceptionV3': tf.keras.applications.InceptionV3,	#311
            'InceptionResNetV2': tf.keras.applications.InceptionResNetV2,
            'MobileNet': tf.keras.applications.MobileNet	,
            'MobileNetV2': tf.keras.applications.MobileNetV2,	
            'DenseNet121': tf.keras.applications.DenseNet121,	
            'DenseNet169': tf.keras.applications.DenseNet169,	
            'DenseNet201': tf.keras.applications.DenseNet201,	
            'NASNetMobile': tf.keras.applications.NASNetMobile	,
            'NASNetLarge': tf.keras.applications.NASNetLarge,	
            'EfficientNetB0': tf.keras.applications.EfficientNetB0,	
            'EfficientNetB1': tf.keras.applications.EfficientNetB1,	
            'EfficientNetB2': tf.keras.applications.EfficientNetB2,	
            'EfficientNetB3': tf.keras.applications.EfficientNetB3,	
            'EfficientNetB4': tf.keras.applications.EfficientNetB4,	
            'EfficientNetB5': tf.keras.applications.EfficientNetB5,	
            'EfficientNetB6': tf.keras.applications.EfficientNetB6,	
            'EfficientNetB7': tf.keras.applications.EfficientNetB7 #813
            }

        ## 2. Create model

        ### 2.1 Load the pre-trained base model

        if baseModel not in modelDict.keys():
            print('{} is not found or not supported. \n Choose from {}'.format(baseModel, modelDict.keys()))
            return

        # Load InceptionV3 model pre-trained on imagenet
        base_model = modelDict[baseModel](input_shape=self.image_size + (3,), # self.image_size is defined in loadData
                                                       include_top=False,
                                                       weights='imagenet')
        # Freeze the base model
        base_model.trainable = False
          

        ### 2.2 Add preprocessing layers and a classification head to build the model

        # Augmentation
        aug_list = []
        if horizontalFlip: 
            aug_list.append(tf.keras.layers.experimental.preprocessing.RandomFlip('horizontal'))
            print('Horizontal flip applied')
        if verticalFlip: 
            aug_list.append(tf.keras.layers.experimental.preprocessing.RandomFlip('vertical'))
            print('Vertical flip applied')
        if randomRotation>0.0: 
            aug_list.append(tf.keras.layers.experimental.preprocessing.RandomRotation(randomRotation))
            print('Random rotation = {} applied'.format(randomRotation))
        if len(aug_list)>0: data_augmentation = tf.keras.Sequential(aug_list)

        # Pre-processing layer
        inputs = tf.keras.Input(shape=self.image_size + (3,))
        if len(aug_list)>0: 
            x = data_augmentation(inputs) # augment
            x = preprocess_input(x) 
        else: x = preprocess_input(inputs) 


        # Then go into the backbone model
        x = base_model(x)

        # Then go into the classification header
        x = tf.keras.layers.GlobalAveragePooling2D()(x)
        x = tf.keras.layers.Dropout(dropout)(x) # You can change the dropout rate 
        prediction_layer = tf.keras.layers.Dense(len(self.classNames), activation='softmax')
        outputs = prediction_layer(x)

        # Put them together
        self.model = tf.keras.Model(inputs, outputs)

        ### Compile the model
        self.model.compile(optimizer=tf.keras.optimizers.Adam(lr=lr1),
              loss=tf.keras.losses.CategoricalCrossentropy(from_logits=True),
              metrics=['accuracy'])
        self.model.summary()

        ### 3. Train the model
        history = self.model.fit(self.train_ds, epochs=initial_epochs, validation_data=self.val_ds)

        # Plot learning curves

        acc = history.history['accuracy']
        val_acc = history.history['val_accuracy']

        loss = history.history['loss']
        val_loss = history.history['val_loss']

        if plot:
            plt.figure(figsize=(8, 8))
            plt.subplot(2, 1, 1)
            plt.plot(acc, label='Training Accuracy')
            plt.plot(val_acc, label='Validation Accuracy')
            plt.legend(loc='lower right')
            plt.ylabel('Accuracy')
            plt.ylim([min(plt.ylim()),1])
            plt.title('Training and Validation Accuracy')

            plt.subplot(2, 1, 2)
            plt.plot(loss, label='Training Loss')
            plt.plot(val_loss, label='Validation Loss')
            plt.legend(loc='upper right')
            plt.ylabel('Cross Entropy')
            plt.ylim([0,1.0])
            plt.title('Training and Validation Loss')
            plt.xlabel('epoch')
            plt.show()

        ## 4. Fine tuning

        ### 4.1 Un-freeze the top layers of the model

        # Un-freeze the whole base model
        base_model.trainable = True

        # Fine-tune from this layer onwards
        #fine_tune_at = 300 # There are a total of 311 layer

        # Freeze all the layers before the `fine_tune_at` layer
        for layer in base_model.layers[:fine_tune_at]:
          layer.trainable =  False

        ### 4.2 Compile the model

        self.model.compile(optimizer=tf.keras.optimizers.Adam(lr=lr2),
                      loss=tf.keras.losses.CategoricalCrossentropy(from_logits=True),
                      metrics=['accuracy'])
        #model.summary()

        ### 4.3 Continue training the model

        total_epochs =  initial_epochs + fine_tune_epochs
        history_fine = self.model.fit(self.train_ds, epochs=total_epochs, initial_epoch=history.epoch[-1], validation_data=self.val_ds)

        # Plot learning curves

        acc += history_fine.history['accuracy']
        val_acc += history_fine.history['val_accuracy']

        loss += history_fine.history['loss']
        val_loss += history_fine.history['val_loss']

        if plot:
            plt.figure(figsize=(8, 8))
            plt.subplot(2, 1, 1)
            plt.plot(acc, label='Training Accuracy')
            plt.plot(val_acc, label='Validation Accuracy')
            plt.ylim([0.8, 1])
            plt.plot([initial_epochs-1,initial_epochs-1],
                      plt.ylim(), label='Start Fine Tuning')
            plt.legend(loc='lower right')
            plt.title('Training and Validation Accuracy')

            plt.subplot(2, 1, 2)
            plt.plot(loss, label='Training Loss')
            plt.plot(val_loss, label='Validation Loss')
            plt.ylim([0, 1.0])
            plt.plot([initial_epochs-1,initial_epochs-1],
                     plt.ylim(), label='Start Fine Tuning')
            plt.legend(loc='upper right')
            plt.title('Training and Validation Loss')
            plt.xlabel('epoch')
            plt.show()

        ## Evaluate the performance of the model
        # Evaluate the overall performance on the val_ds set
        #loss, accuracy = self.model.evaluate(self.val_ds)
        #print('Val accuracy :', accuracy)

        ## 5. Save model

        # Save the model in the current folder 
        self.model.save(self.modelFile) 
        print('Model saved at ', self.modelFile)

        modelDetails = {
                'modelName' : self.modelName,
                'classNames' : self.classNames
            }
        with open(self.modelDetailFile, 'w') as outfile:
            json.dump(modelDetails, outfile)
        print('Model details saved in ', self.modelDetailFile)

    def evaluate(self):
        prediction_tensor = self.model.predict(self.val_ds)
        prediction_int = list(np.argmax(prediction_tensor,axis=1))
        labels_int = []
        for i,gt in self.val_ds:
          for j in gt:
            labels_int.append(np.argmax(j))

        class_names = materialClassifier.classNames
        prediction = [class_names[i] for i in prediction_int]
        label = [class_names[i] for i in labels_int]

        print(f'Accuracy is   : {accuracy_score(prediction,label)}')
        print(f'F1 score is   : {accuracy_score(f1_score,label)}')
        cnf_matrix = confusion_matrix(prediction,label)
        plot_confusion_matrix(cnf_matrix, classes=class_names, title='Confusion matrix',normalize=True,xlabel='Labels',ylabel='Predictions')



    def retrain(self,lr1=0.0001,initial_epochs=10):
        '''
        if self.train_ds.class_names != self.classNames:
            print('Can not retrain. Folder names mismatch predefined classNames: {}'.format(self.classNames))
            return
        '''

        ### Compile the model
        self.model.compile(optimizer=tf.keras.optimizers.Adam(lr=lr1),
              loss=tf.keras.losses.CategoricalCrossentropy(from_logits=True),
              metrics=['accuracy'])
        self.model.summary()

        ### Train the model
        history = self.model.fit(self.train_ds, epochs=initial_epochs, validation_data=self.val_ds)

    def save(self, newModelName=None):
        if newModelName != None: 
            newFileName = os.path.join(self.workDir,'{}.h5'.format(newModelName))
        else:
            newFileName = self.modelFile
        self.model.save(newFileName) 
        print('Model saved at ', newFileName)



def main():
    pass

if __name__ == '__main__':
    main()







from time import sleep

import torch
from PIL import Image
from torchvision import transforms
import numpy as np
import os

from .csail_seg.config import cfg
from .csail_seg.models import ModelBuilder
from scipy.io import loadmat
torch.multiprocessing.set_start_method('spawn', force=True)
normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                 std=[0.229, 0.224, 0.225])

cur_dir_tmp = os.path.dirname(os.path.realpath(__file__))

class MaskBuilding(object):
    #colors = loadmat('./csail_segmentation_tool/csail_seg/data/color150.mat')['colors']
    #config_file = './csail_segmentation_tool/csail_seg/config/ade20k-resnet50dilated-ppm_deepsup.yaml'
    colors = loadmat(f'{cur_dir_tmp}/csail_seg/data/color150.mat')['colors']
    config_file = f'{cur_dir_tmp}/csail_seg/config/ade20k-resnet50dilated-ppm_deepsup.yaml'
    cfg.merge_from_file(config_file)
    cfg.MODEL.arch_encoder = cfg.MODEL.arch_encoder.lower()
    cfg.MODEL.arch_decoder = cfg.MODEL.arch_decoder.lower()
    enc_weights = f'{cur_dir_tmp}/csail_seg/ade20k-resnet50dilated-ppm_deepsup/encoder_epoch_20.pth'
    dec_weights = f'{cur_dir_tmp}/csail_seg/ade20k-resnet50dilated-ppm_deepsup/decoder_epoch_20.pth'
    cfg.MODEL.weights_encoder = enc_weights
    cfg.MODEL.weights_decoder = dec_weights

    def round2nearest_multiple(self, x, p):
        return ((x - 1) // p + 1) * p

    def __init__(self, device, model_dir=''):

        if model_dir!='':
            enc_weights = f'{model_dir}/encoder_epoch_20.pth'
            dec_weights = f'{model_dir}/decoder_epoch_20.pth'
            cfg.MODEL.weights_encoder = enc_weights
            cfg.MODEL.weights_decoder = dec_weights

        self.device = device
        self.net_encoder = ModelBuilder.build_encoder(
            arch=cfg.MODEL.arch_encoder,
            fc_dim=cfg.MODEL.fc_dim,
            weights=cfg.MODEL.weights_encoder)
        self.net_decoder = ModelBuilder.build_decoder(
            arch=cfg.MODEL.arch_decoder,
            fc_dim=cfg.MODEL.fc_dim,
            num_class=cfg.DATASET.num_class,
            weights=cfg.MODEL.weights_decoder,
            use_softmax=True)

        self.net_encoder.eval().to(self.device)
        self.net_decoder.eval().to(self.device)

    def get_mask(self,pic):

        ori_height = pic.size[0]
        ori_width = pic.size[1]
        print('Calculating mask for image')
        img_resized_list = []
        with torch.no_grad():
            for this_short_size in cfg.DATASET.imgSizes:
                # calculate target height and width
                scale = min(this_short_size / float(min(ori_height, ori_width)),
                            cfg.DATASET.imgMaxSize / float(max(ori_height, ori_width)))
                target_height, target_width = int(ori_height * scale), int(ori_width * scale)

                # to avoid rounding in network
                target_width = self.round2nearest_multiple(target_width, cfg.DATASET.padding_constant)
                target_height = self.round2nearest_multiple(target_height, cfg.DATASET.padding_constant)

                # resize and normalize images
                val_transform = transforms.Compose(
                    [transforms.Resize((target_width, target_height)), transforms.ToTensor(), normalize])

                # image transform, to torch float tensor 3xHxW
                img_resized = val_transform(pic)
                img_resized = torch.unsqueeze(img_resized.to(self.device), 0)
                img_resized_list.append(img_resized)

            scores = torch.zeros(img_resized_list[0].shape[0], cfg.DATASET.num_class, ori_width, ori_height).to(self.device)

            for img in img_resized_list:
                # forward pass
                pred_tmp = self.net_decoder(self.net_encoder(img, return_feature_maps=True),
                                            segSize=(ori_width, ori_height))
                scores = scores + pred_tmp / len(cfg.DATASET.imgSizes)

        _, segm_pred = torch.max(scores, dim=1)

        return segm_pred.cpu()

    def __call__(self, pic):
        """
        Args:
            pic (PIL Image): Image with building and surrounds
        Returns:
            Tensor: Cut out building with everything else black
        """
        segm_pred = self.get_mask(pic)

        image = np.array(pic)

        mask = np.array(segm_pred.cpu()).squeeze()
        # Filter building labels which are 25 and 1
        mask[np.where((mask != 25) & (mask != 1))] = 0
        image[mask == 0, :] = 0

        image = Image.fromarray(np.uint8(image))

        return image

    def __repr__(self):
        return self.__class__.__name__ + '()'



import torch


class Sampler(object):
    """Base class for all Samplers.

    Every Sampler subclass has to provide an __iter__ method, providing a way
    to iterate over indices of dataset elements, and a __len__ method that
    returns the length of the returned iterators.
    """

    def __init__(self, data_source):
        pass

    def __iter__(self):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError


class SequentialSampler(Sampler):
    """Samples elements sequentially, always in the same order.

    Arguments:
        data_source (Dataset): dataset to sample from
    """

    def __init__(self, data_source):
        self.data_source = data_source

    def __iter__(self):
        return iter(range(len(self.data_source)))

    def __len__(self):
        return len(self.data_source)


class RandomSampler(Sampler):
    """Samples elements randomly, without replacement.

    Arguments:
        data_source (Dataset): dataset to sample from
    """

    def __init__(self, data_source):
        self.data_source = data_source

    def __iter__(self):
        return iter(torch.randperm(len(self.data_source)).long())

    def __len__(self):
        return len(self.data_source)


class SubsetRandomSampler(Sampler):
    """Samples elements randomly from a given list of indices, without replacement.

    Arguments:
        indices (list): a list of indices
    """

    def __init__(self, indices):
        self.indices = indices

    def __iter__(self):
        return (self.indices[i] for i in torch.randperm(len(self.indices)))

    def __len__(self):
        return len(self.indices)


class WeightedRandomSampler(Sampler):
    """Samples elements from [0,..,len(weights)-1] with given probabilities (weights).

    Arguments:
        weights (list)   : a list of weights, not necessary summing up to one
        num_samples (int): number of samples to draw
        replacement (bool): if ``True``, samples are drawn with replacement.
            If not, they are drawn without replacement, which means that when a
            sample index is drawn for a row, it cannot be drawn again for that row.
    """

    def __init__(self, weights, num_samples, replacement=True):
        self.weights = torch.DoubleTensor(weights)
        self.num_samples = num_samples
        self.replacement = replacement

    def __iter__(self):
        return iter(torch.multinomial(self.weights, self.num_samples, self.replacement))

    def __len__(self):
        return self.num_samples


class BatchSampler(object):
    """Wraps another sampler to yield a mini-batch of indices.

    Args:
        sampler (Sampler): Base sampler.
        batch_size (int): Size of mini-batch.
        drop_last (bool): If ``True``, the sampler will drop the last batch if
            its size would be less than ``batch_size``

    Example:
        >>> list(BatchSampler(range(10), batch_size=3, drop_last=False))
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
        >>> list(BatchSampler(range(10), batch_size=3, drop_last=True))
        [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
    """

    def __init__(self, sampler, batch_size, drop_last):
        self.sampler = sampler
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __iter__(self):
        batch = []
        for idx in self.sampler:
            batch.append(idx)
            if len(batch) == self.batch_size:
                yield batch
                batch = []
        if len(batch) > 0 and not self.drop_last:
            yield batch

    def __len__(self):
        if self.drop_last:
            return len(self.sampler) // self.batch_size
        else:
            return (len(self.sampler) + self.batch_size - 1) // self.batch_size



import torch
from PIL import Image
from torch.utils.data import Dataset

import numpy as np
import os
from collections import defaultdict
from scipy import signal

class YearBuiltFolder(Dataset):
    
    def __init__(self, image_folder, soft_labels=False, gaussian_std=1.5,  transforms=None, classes=None, calc_perf=False):

        self.transforms = transforms

        self.img_paths = []
        self.filenames = []
        self.labels = []
        self.calc_perf = calc_perf
        self.soft_labels = soft_labels
        '''
        if not os.path.isdir(image_folder):
            if os.path.isfile(image_folder):
                # The following format is to be consistent with os.walk output
                file_list = [[os.path.split(image_folder)[0],'None',[os.path.split(image_folder)[1]]]]
            else:
                print('Error: Image folder or file {} not found.'.format(image_folder))
                exit()
        else:
            file_list = os.walk(image_folder, followlinks=True)
        '''

        if isinstance(image_folder, list): #a list of images
            file_list = [[os.path.split(i)[0],'None',[os.path.split(i)[1]]] for i in image_folder]
        elif isinstance(image_folder, str): 
            if not os.path.isdir(image_folder):
                if os.path.isfile(image_folder):
                    # The following format is to be consistent with os.walk output
                    file_list = [[os.path.split(image_folder)[0],'None',[os.path.split(image_folder)[1]]]]
                else:
                    print('Error: Image folder or file {} not found.'.format(image_folder))
                    exit()

            else:# dir
                file_list = os.walk(image_folder, followlinks=True)

        class_counts = defaultdict(lambda:0) 

        for root, _, fnames in sorted(file_list):
            for fname in sorted(fnames):
                
                if 'jpg' in fname or 'png' in fname:
                    img_path = os.path.join(root, fname)

                    self.filenames.append(fname)
                    self.img_paths.append(img_path)
                    
                    if calc_perf:
                        label = root.split(os.path.sep)[-1]
                        self.labels.append(label)
                        class_counts[label] = class_counts[label] + 1 # count labels
        
        if classes is None:
            self.classes = np.unique(self.labels)
        else:
            self.classes=classes
        ########### Calculate train weights for weighted sampling        
        self.class_weights = {}
        self.train_weights = []

        for _class in self.classes:
            self.class_weights[_class] = sum(np.array([label == _class for label in self.labels]))
        
        for entry in self.labels:
            self.train_weights.append(1 / self.class_weights[entry])       

        ################################
        # Optionally, create soft labels by pre-calculating unimodal gaussians and assigning
        # them according to the class label
            
        if soft_labels:
            res = 100
            class_n = len(self.classes)
            window = signal.gaussian(res*class_n, std=(res/10)*(class_n-1)*float(gaussian_std))
            
            center_id = int((res*class_n)/2)
            
            samples = []
            for i in range(-class_n//2,class_n//2+1):
                pos_of_int = window[min(max(center_id+res*i,0),window.shape[0]-1)]
                samples.append(pos_of_int)
                
            samples = np.array(samples)
        
            label_lookup = defaultdict()
            for class_id in self.classes:
                class_soft_labels = []
                
                max_class_id = class_n
                for i in range(max_class_id):
                    sample_id = (i+3)-np.flatnonzero(self.classes == class_id)
            
                    if sample_id < 0 or sample_id >= max_class_id:
                        class_soft_labels.append(0.0)
                    else:
                        class_soft_labels.append(samples[sample_id].squeeze())
                        
                class_soft_labels = np.array(class_soft_labels)
                
                label_lookup[class_id] = class_soft_labels
        
            year_built_softlabels = [label_lookup[label] for label in self.labels]
            
            self.labels = year_built_softlabels
            
            

    def __len__(self):
        return len(self.filenames)

    def loader(self, path):
        with open(path, 'rb') as f:
            img = Image.open(f)
            return img.convert('RGB')
        
    def __getitem__(self, index):

        
        path = self.img_paths[index]
        img = self.loader(path)

        if self.transforms is not None:
            img = self.transforms(img)
        
        if self.calc_perf:
            label = self.labels[index]
            # The label is either a probability distribution or the class number
            if self.soft_labels:
                target = label
                target = torch.FloatTensor(target) 
            else:
                target = [np.flatnonzero(self.classes == label)]
                target = torch.LongTensor(target).squeeze()
                
            
        else:
            target = []
    

        return img, target, path

  
    
